"""
(C) Copyright 2019 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on Apr 25, 2018


A module implementing several doubly-robust methods.
These methods utilize causal standardization model and causal weight models and combine them to hopefully achieve a
better model.
The exact way to combine differs from different models and is described in the class docstring.
"""

import abc
import warnings

import pandas as pd
import numpy as np

from .base_estimator import IndividualOutcomeEstimator
from .base_weight import WeightEstimator, PropensityEstimator
from ..utils import general_tools as g_tools
from ..utils.stat_utils import robust_lookup


class BaseDoublyRobust(IndividualOutcomeEstimator):
    """
    Abstract class defining the interface and general initialization of specific doubly-robust methods.
    """

    def __init__(self, outcome_model, weight_model,
                 outcome_covariates=None, weight_covariates=None):
        """

        Args:
            outcome_model(IndividualOutcomeEstimator): A causal model that estimate on individuals level
                                                      (e.g. Standardization).
            weight_model (WeightEstimator): A causal model for weighting individuals (e.g. IPW).
            outcome_covariates (array): Covariates to use for outcome model.
                                        If None - all covariates passed will be used.
                                        Either list of column names or boolean mask.
            weight_covariates (array): Covariates to use for weight model.
                                       If None - all covariates passed will be used.
                                       Either list of column names or boolean mask.
        """
        super(BaseDoublyRobust, self).__init__(lambda **x: None)  # Dummy initialization
        delattr(self, "learner")  # To remove the learner attribute a IndividualOutcomeEstimator has
        self.outcome_model = outcome_model
        self.weight_model = weight_model
        self.outcome_covariates = outcome_covariates
        self.weight_covariates = weight_covariates

    @abc.abstractmethod
    def fit(self, X, a, y, refit_weight_model=True, **kwargs):
        raise NotImplementedError

    def _extract_outcome_model_data(self, X):
        outcome_covariates = self.outcome_covariates or X.columns
        X_outcome = X[outcome_covariates]
        return X_outcome

    def _extract_weight_model_data(self, X):
        weight_covariates = self.weight_covariates or X.columns
        X_weight = X[weight_covariates]
        return X_weight

    def _prepare_data(self, X, a):
        """
        Extract the relevant parts for outcome model and weight model for the entire data matrix

        Args:
            X (pd.DataFrame): Covariate matrix of size (num_subjects, num_features).
            a (pd.Series): Treatment assignment of size (num_subjects,).

        Returns:
            (pd.DataFrame, pd.DataFrame): X_outcome, X_weight
                Data matrix for outcome model and data matrix weight model
        """
        X_outcome = self._extract_outcome_model_data(X)
        X_weight = self._extract_weight_model_data(X)
        return X_outcome, X_weight

    def _is_weight_model_fitted(self):
        try:
            weight_model_is_fitted = g_tools.check_learner_is_fitted(self.weight_model.learner)
        except AttributeError:  # `weight_model` has no `learner` attribute. fit anyway
            warnings.warn(
                f"Weight model of type {type(self.weight_model)} does not"
                f"have a `learner` attribute and cannot be tested whether"
                f"it is fitted or not. Attempting to fit anyway."
            )
            weight_model_is_fitted = False
        return weight_model_is_fitted

    def __repr__(self):
        repr_string = g_tools.create_repr_string(self)
        # Make a new line between outcome_model and weight_model
        repr_string = repr_string.replace(", weight_model",
                                          ",\n{spaces}weight_model".format(spaces=" " * (len(self.__class__.__name__)
                                                                                         + 1)))
        return repr_string


class AIPW(BaseDoublyRobust):
    def __init__(self, outcome_model, weight_model,
                 outcome_covariates=None, weight_covariates=None,
                 overlap_weighting=False):
        """
        Calculates a doubly-robust estimate of the treatment effect by performing
        potential-outcome prediction (`outcome_model`) and then correcting its
        prediction-residuals using re-weighting from a treatment model (`weight_model`, like IPW).

        It has two flavors, which slightly change the weighting of the outcome model in the correction term.
        Let e(X) be the estimated propensity score and m(X, A) is the estimated effect by an estimator,
        then the individual estimates are:

        | m(X,1) + A*(Y-m(X,1))/e(X), and
        | m(X,0) + (1-A)*(Y-m(X,0))/(1-e(X))

        Which are basically add IP-weighted residuals from the observed predictions.
        As described in Kang and Schafer (2007) section 3.1 and Robins, Rotnitzky, and Zhao (1994).

        The additional flavor when `overlap_weighting=True` is from Glynn and Quinn (2010),
        adds weighting by the propensity-of-the-other-class to the outcome model,
        so extreme example (with poor covariate overlap) will contribute less to the correction
        (i.e. rely less on their prediction value that might extrapolate too much).
        This is a similar notion used in Overlap Weights model (hence the argument name)

        | A * [Y - (1-e(X))m(X,1)]/e(X) + (1-A) * m(X,1), and
        | (1-A) * [Y - e(X)m(X,0)]/(1-e(X)) + A * m(X,0)

        Args:
            outcome_model(IndividualOutcomeEstimator): A causal model that estimate on individuals level
                                                      (e.g. Standardization).
            weight_model (WeightEstimator | PropensityEstimator): A causal model for weighting individuals (e.g. IPW).
                If `overlap_weighting=True` then must be a `PropensityEstimator` model.
            outcome_covariates (array): Covariates to use for outcome model.
                If None - all covariates passed will be used. Either list of column names or boolean mask.
            weight_covariates (array): Covariates to use for weight model.
                If None - all covariates passed will be used. Either list of column names or boolean mask.
            overlap_weighting (bool): Whether to tweak the outcome-model correction-term to rely less on
                data-points with poor covariate overlap (extreme propensity).
                if `True`, requires `weight_model` to be an instance of `PropensityEstimator`.

        References:
            * Kang and Schafer, 2007, (https://dx.doi.org/10.1214/07-STS227)
            * Kang and Schafer attribute the original method to Cassel, Särndal and Wretman.
            * Glynn and Quinn, 2010, https://doi.org/10.1093/pan/mpp036
            * Robins, Rotnitzky, and Zhao, 1994, https://doi.org/10.1080/01621459.1994.10476818
        """
        super().__init__(outcome_model, weight_model,
                         outcome_covariates, weight_covariates)
        self.overlap_weighting = overlap_weighting

    def fit(self, X, a, y, refit_weight_model=True, **kwargs):
        if self.overlap_weighting and a.nunique() != 2:
            raise AssertionError(
                f"`overlap_weights=True` version can only be used with binary treatment."
                f"Instead, treatment values are {set(a)}. Try setting it to `False`."
            )

        X_outcome, X_weight = self._prepare_data(X, a)
        weight_model_is_not_fitted = not self._is_weight_model_fitted()

        if refit_weight_model or weight_model_is_not_fitted:
            self.weight_model.fit(X=X_weight, a=a, y=y)

        self.outcome_model.fit(X=X_outcome, y=y, a=a)
        return self

    # def predict(self, X, a):
    #     raise NotImplementedError("Predict is not well defined for doubly robust and thus unimplemented.")

    def estimate_individual_outcome(self, X, a, treatment_values=None, predict_proba=None):
        """
        Estimates individual outcome under different treatment values (interventions).

        Notes:
            This method utilizes only the standardization model behind the doubly-robust model. Namely, this is an
            uncorrected outcome (that does not incorporates the weighted observed outcome).
            To get a true doubly-robust estimation use the estimate_population_outcome, rather than an individual
            outcome.

        Args:
            X (pd.DataFrame): Covariate matrix of size (num_subjects, num_features).
            a (pd.Series): Treatment assignment of size (num_subjects,).
            treatment_values (Any): Desired treatment value/s to use when estimating the counterfactual outcome/
                                    If not supplied, calculates for all available treatment values.
            predict_proba (bool | None): In case the outcome task is classification and in case `learner` supports the
                                         operation, if True - prediction will utilize learner's `predict_proba` or
                                         `decision_function` which returns a continuous matrix of size
                                         (n_samples, n_classes).
                                         If False - `predict` will be used and return value will be based on a vector
                                         of class classifications.

        Returns:
            pd.DataFrame: DataFrame which columns are treatment values and rows are individuals: each column is a vector
                          size (num_samples,) that contains the estimated outcome for each individual under the
                          treatment value in the corresponding key.
        """
        X_outcome = self._extract_outcome_model_data(X)
        prediction = self.outcome_model.estimate_individual_outcome(X_outcome, a, treatment_values, predict_proba)
        return prediction

    def _estimate_corrected_individual_outcome(self, X, a, y, treatment_values=None, predict_proba=None):
        """
        Estimating corrected individual counterfactual outcomes.
        This correction is an intermediate stage needed to estimate doubly-robust population outcome.

        Args:
            X (pd.DataFrame): Covariate matrix of size (num_subjects, num_features).
            a (pd.Series): Treatment assignment of size (num_subjects,).
            y (pd.Series): Observed outcome of size (num_subjects,).
            treatment_values (Any): Desired treatment value/s to use when estimating the counterfactual outcome.
                                    If not supplied, calculates for all available treatment values.
            predict_proba (bool | None): In case the outcome task is classification and in case `learner` supports the
                                         operation, if True - prediction will utilize learner's `predict_proba` or
                                         `decision_function` which returns a continuous matrix of size
                                         (n_samples, n_classes).
                                         If False - `predict` will be used and return value will be based on a vector
                                         of class classifications.

        Returns:
            pd.DataFrame: DataFrame which columns are treatment values and rows are individuals: each column is a vector
                          size (num_samples,) that contains the estimated outcome for each individual under the
                          treatment value in the corresponding key.
        """
        X_outcome, X_weight = self._prepare_data(X, a)
        weights = self.weight_model.compute_weights(X_weight, a)
        individual_cf = self.estimate_individual_outcome(X_outcome, a, treatment_values, predict_proba)
        is_predict_proba_classification_result = isinstance(individual_cf.columns, pd.MultiIndex)
        if is_predict_proba_classification_result:
            # Classification `outcome_model` with `predict_proba=True` returns a MultiIndex of treatments over outcomes.
            # Extract the prediction for the maximal outcome class (probably class `1` in binary classification):
            outcome_values = individual_cf.columns.get_level_values(level=-1)
            individual_cf = individual_cf.xs(
                outcome_values.max(), axis="columns", level=-1, drop_level=True,
            )
        factual_prediction = robust_lookup(individual_cf, a)

        if self.overlap_weighting:
            propensities = self.weight_model.compute_propensity_matrix(X_weight)
            reversed_propensities = robust_lookup(propensities, 1 - a)  # take propensities of opposite treatment group
            factual_prediction *= reversed_propensities
            corrected_outcome = (y - factual_prediction) * weights
        else:
            outcome_correction = (y - factual_prediction) * weights
            corrected_outcome = factual_prediction + outcome_correction

        for treatment_value in treatment_values:
            individual_cf.loc[a == treatment_value, treatment_value] = corrected_outcome.loc[a == treatment_value]

        return individual_cf

    def estimate_population_outcome(self, X, a, y=None, treatment_values=None, predict_proba=None, agg_func="mean"):
        """
        Doubly-robust averaging, combining the individual counterfactual predictions from the standardization model
        and the weighted observed outcomes to estimate population outcome for each treatment subgroup.

        Args:
            X (pd.DataFrame): Covariate matrix of size (num_subjects, num_features).
            a (pd.Series): Treatment assignment of size (num_subjects,).
            y (pd.Series): Observed outcome of size (num_subjects,).
            treatment_values (Any): Desired treatment value/s to stratify upon before aggregating individual into
                                    population outcome.
                                    If not supplied, calculates for all available treatment values.
            predict_proba (bool | None): To be used when provoking estimate_individual_outcome.
                                         In case the outcome task is classification and in case `learner` supports the
                                         operation, if True - prediction will utilize learner's `predict_proba` or
                                         `decision_function` which returns a continuous matrix of size
                                         (n_samples, n_classes).
                                         If False - `predict` will be used and return value will be based on a vector
                                         of class classifications.
            agg_func (str): Type of aggregation function (e.g. "mean" or "median").

        Returns:
            pd.Series: Series which index are treatment values, and the values are numbers - the aggregated outcome for
                       the strata of people whose assigned treatment is the key.

        """
        if y is None:
            raise TypeError("Must supply outcome variable (y). Got None instead.")
        treatment_values = g_tools.get_iterable_treatment_values(treatment_values, a)

        individual_cf = self._estimate_corrected_individual_outcome(X, a, y, treatment_values)

        population_outcome = individual_cf.apply(self._aggregate_population_outcome, args=(agg_func,))
        return population_outcome

    def estimate_effect(self, outcome1, outcome2, agg="population", effect_types="diff"):
        if isinstance(outcome1, pd.Series) and isinstance(outcome2, pd.Series) and agg == "population":
            warnings.warn("Seems you're trying to calculate population (average) effect from individual outcome "
                          "prediction.\n"
                          "Note that the result might be biased since the output of estimate_individual_outcome() is "
                          "not corrected for population effect.\n"
                          "In case you want individual effect use agg='individual', or in case you want population"
                          "effect use the estimate_population_effect() output as your input to this function.")
        effect = super().estimate_effect(outcome1, outcome2, agg, effect_types)
        return effect


class PropensityFeatureStandardization(BaseDoublyRobust):
    def __init__(self, outcome_model, weight_model,
                 outcome_covariates=None, weight_covariates=None,
                 feature_type="weight_vector"):
        """
        A doubly-robust estimator of the effect of treatment.
        This model adds the weighting (inverse probability weighting)
        as additional feature to the outcome model.

        References:
            * Bang and Robins, https://doi.org/10.1111/j.1541-0420.2005.00377.x
            * Kang and Schafer, section 3.3, https://dx.doi.org/10.1214/07-STS227

        Args:
            outcome_model(IndividualOutcomeEstimator): A causal model that estimate on individuals level
            weight_model (WeightEstimator | PropensityEstimator): A causal model for weighting individuals (e.g. IPW).
            outcome_covariates (array): Covariates to use for outcome model.
                If None - all covariates passed will be used. Either list of column names or boolean mask.
            weight_covariates (array): Covariates to use for weight model.
                If None - all covariates passed will be used. Either list of column names or boolean mask.
            feature_type (str): the type of covariate to add. One of the following options:
                *  "weight_vector": uses a signed weight vector. Only defined for binary treatment.
                   For example, if `weight_model` is IPW then: 1/Pr[A=a_i|X] for each sample `i`.
                   As described in Bang and Robins (2005).
                * "signed_weight_vector": as `'weight_vector'`, but negates the weights of the control group.
                  For example, if `weight_model` is IPW then: 1/Pr[A|X] for treated and 1/Pr[A|X] for controls.
                  As described in the correction for Bang and Robins (2008)
                * "weight_matrix": uses the entire weight matrix.
                   For example, if `weight_model` is IPW then: 1/Pr[A_i=a|X_i=x],
                                for all treatment values `a` and for every sample `i`.
                * "masked_weighted_matrix": uses the entire weight matrix, but masks it with a dummy-encoding
                  of the treatment assignment.
                  For example, if weight_model` is IPW then: 1/Pr[A=a_i|X=x_i] and 0 for all other `a≠a_i` columns.
                  As described in Bang and Robins (2005).
                * "propensity_vector": uses the probabilities for being in treatment group: Pr[A=1|X].
                                       Better defined for binary treatment.
                                       Equivalent to Scharfstein, Rotnitzky, and Robins (1999) that use its inverse.
                * "logit_propensity_vector": uses logit transformation of the propensity to treat Pr[A=1|X].
                                             As described in Kang and Schafer (2007)
                * "propensity_matrix": uses the probabilities for all treatment options,
                    Pr[A_i=a|X_i=x] for all treatment values `a` and samples `i`.
        """
        super().__init__(outcome_model, weight_model,
                         outcome_covariates, weight_covariates)
        self.feature_type = feature_type

    def estimate_individual_outcome(self, X, a, treatment_values=None, predict_proba=None):
        X_augmented = self._augment_outcome_model_data(X, a)
        prediction = self.outcome_model.estimate_individual_outcome(X_augmented, a, treatment_values, predict_proba)
        return prediction

    def _augment_outcome_model_data(self, X, a):
        """
        This method adds the features needed to the model (either weight vector or weight matrix, depending on the
        underlining weight_estimator).

        Args:
            X (pd.DataFrame): Covariate matrix of size (num_subjects, num_features).
            a (pd.Series): Treatment assignment of size (num_subjects,).

        Returns:
            pd.DataFrame: Concatenation of the weight features (either single or multi column) to the provided covariate
                          matrix (W | X).
        """
        X_outcome, X_weight = self._prepare_data(X, a)
        feature_func = self._get_feature_function(self.feature_type)
        weights_feature = feature_func(X_weight, a)
        # Let standardization deal with incorporating treatment assignment (a) into the data:
        X_augmented = pd.concat([weights_feature, X_outcome], join="outer", axis="columns")
        return X_augmented

    # def predict(self, X, a):
    #     raise NotImplementedError

    def fit(self, X, a, y, refit_weight_model=True, **kwargs):
        X_outcome, X_weight = self._prepare_data(X, a)
        weight_model_is_not_fitted = not self._is_weight_model_fitted()

        if refit_weight_model or weight_model_is_not_fitted:
            self.weight_model.fit(X=X_weight, a=a, y=y)

        X_augmented = self._augment_outcome_model_data(X, a)
        self.outcome_model.fit(X=X_augmented, y=y, a=a)
        return self

    def _get_feature_function(self, function_name):

        def weight_vector(X, a):
            w = self.weight_model.compute_weights(X, a)
            w = w.rename("ipf")
            return w

        def signed_weight_vector(X, a):
            """Bang and Robins, 2005 / 2008"""
            if a.nunique() != 2:
                raise AssertionError(
                    f"`feature_type` 'weight_vector' can only be used with binary treatment."
                    f"Instead, treatment values are {set(a)}."
                )
            w = weight_vector(X, a)
            w[a == 0] *= -1
            return w

        def weight_matrix(X, a):
            W = self.weight_model.compute_weight_matrix(X, a)
            W = W.add_prefix("ipf_")
            return W

        def masked_weight_matrix(X, a):
            """Bang and Robins, 2005"""
            W = weight_matrix(X, a)
            A = pd.get_dummies(a)
            A = A.add_prefix("ipf_")  # To match naming of `W`
            W_masked = W * A
            return W_masked

        def propensity_vector(X, a):
            p = self.weight_model.compute_propensity(X, a)
            p = p.rename("propensity")
            return p

        def logit_propensity_vector(X, a, safe=True):
            p = propensity_vector(X, a)
            if safe:
                epsilon = np.finfo(float).eps
                p = np.clip(p, epsilon, 1 - epsilon)
            return np.log(p / (1 - p))

        def propensity_matrix(X, a):
            P = self.weight_model.compute_propensity_matrix(X)
            # P = P.iloc[:, 1:]  # Drop first column
            P = P.add_prefix("propensity_")
            return P

        feature_functions = {
            "weight_vector": weight_vector,
            "signed_weight_vector": signed_weight_vector,
            "weight_matrix": weight_matrix,
            "masked_weight_matrix": masked_weight_matrix,
            "propensity_vector": propensity_vector,
            "logit_propensity_vector": logit_propensity_vector,
            "propensity_matrix": propensity_matrix,
        }
        return feature_functions[function_name]


class WeightedStandardization(BaseDoublyRobust):
    """
    This model uses the weights from the weight-model (e.g. inverse probability weighting)
    as individual weights for fitting the outcome model.

    References:
        * Kang and Schafer, section 3.2, https://dx.doi.org/10.1214/07-STS227
    """

    def estimate_individual_outcome(self, X, a, treatment_values=None, predict_proba=None):
        prediction = self.outcome_model.estimate_individual_outcome(X=X, a=a, treatment_values=treatment_values,
                                                                    predict_proba=predict_proba)
        return prediction

    def fit(self, X, a, y, refit_weight_model=True, **kwargs):
        X_outcome, X_weight = self._prepare_data(X, a)
        weight_model_is_not_fitted = not self._is_weight_model_fitted()

        if refit_weight_model or weight_model_is_not_fitted:
            self.weight_model.fit(X=X_weight, a=a, y=y)

        weights = self.weight_model.compute_weights(X_weight, a)
        self.outcome_model.fit(X=X_outcome, y=y, a=a, sample_weight=weights)
        return self

    # def predict(self, X, a):
    #     prediction = self.outcome_model.predict(X, a)
    #     return prediction



import numpy as np
import unittest
import pandas as pd
from sklearn.linear_model import LogisticRegression
from sklearn.preprocessing import PolynomialFeatures
from sklearn.pipeline import Pipeline
from causallib.datasets.data_loader import load_nhefs_survival
from causallib.survival.survival_utils import get_person_time_df, safe_join
from causallib.estimation.ipw import IPW
from causallib.survival.standardized_survival import StandardizedSurvival
from causallib.survival.weighted_standardized_survival import WeightedStandardizedSurvival
from causallib.survival.weighted_survival import WeightedSurvival
from causallib.survival.marginal_survival import MarginalSurvival
try:
    import lifelines
    LIFELINES_FOUND = True
except ImportError:
    LIFELINES_FOUND = False

"""
Simulated Test Data For Drug Effects
-------------------------------------
Drug A is simulated with a beneficial effect, drug B with zero effect.
x_0 is a dummy variable that has no effect on treatment nor on the outcome.
Drug C involves informative censoring, so don't expect accurate results
with censoring-naive models.

Data generation and fitting yielded the following:
  IPW-LR S diff IPW-LR effect S-LR S diff S-LR effect T-LR S diff T-LR effect  oracle effect
A          0.24          0.45        0.31        0.43        0.34        0.49           0.41
B         -0.39        -0.077       -0.38       -0.01       -0.39     -0.0069          -0.06
C          0.34           0.5        0.27        0.26         0.3        0.35           0.56

Where -
* S diff = difference in observed survival curves at end of follow-up
* effect = difference in adjusted survival curves at the end of follow-up
* IPW-LR = inverse propensity weighting with logistic regression
* S-LR = standardization with logistic regression (non-stratified)

"""

RANDOM_SEED = 1
TEST_DATA_TTE_DRUG_EFFECTS = dict()
TEST_DATA_TTE_DRUG_EFFECTS['A'] = pd.DataFrame(
    columns=['x_0', 'x_1', 'a', 'y', 't'],
    data=[
        [0, 0, 0, 0, 1],
        [1, 1, 1, 1, 1],
        [0, 0, 1, 0, 3],
        [0, 0, 0, 1, 2],
        [0, 0, 1, 0, 2],
        [0, 0, 0, 1, 2],
        [0, 1, 1, 0, 2],
        [0, 0, 1, 0, 3],
        [0, 1, 1, 1, 2],
        [1, 1, 0, 1, 1],
        [0, 0, 0, 1, 2],
        [1, 0, 0, 0, 1],
        [0, 1, 0, 1, 1],
        [1, 0, 0, 0, 2],
        [0, 0, 0, 0, 3],
        [1, 1, 0, 1, 1],
        [0, 1, 0, 1, 1],
        [1, 1, 0, 1, 1],
        [0, 1, 1, 1, 2],
        [0, 1, 1, 1, 3],
        [1, 1, 1, 1, 2],
        [1, 0, 0, 1, 1],
        [0, 0, 0, 0, 2],
        [1, 1, 1, 1, 1],
        [1, 0, 1, 0, 1],
        [1, 0, 0, 0, 1],
        [0, 1, 1, 0, 1],
        [0, 0, 1, 0, 2],
        [0, 1, 1, 0, 1],
        [1, 1, 1, 0, 2],
        [0, 1, 1, 0, 2],
        [0, 1, 0, 1, 1],
        [1, 1, 1, 1, 2],
        [1, 0, 0, 1, 2],
        [1, 0, 1, 0, 3],
        [0, 1, 1, 1, 2],
        [1, 0, 1, 0, 2],
        [1, 1, 1, 1, 2],
        [0, 1, 0, 1, 1],
        [1, 1, 0, 1, 1],
        [1, 0, 0, 0, 2],
        [1, 1, 1, 0, 3],
        [0, 0, 0, 1, 2],
        [1, 1, 1, 1, 1],
        [0, 0, 0, 1, 1],
        [0, 0, 0, 1, 1],
        [1, 1, 1, 1, 1],
        [0, 1, 1, 0, 3],
        [0, 0, 0, 0, 2],
        [0, 1, 1, 1, 1]
    ]
)
TEST_DATA_TTE_DRUG_EFFECTS['B'] = pd.DataFrame(
    columns=['x_0', 'x_1', 'a', 'y', 't'],
    data=[
        [0, 0, 0, 0, 3],
        [1, 1, 0, 1, 1],
        [0, 0, 1, 0, 3],
        [0, 0, 1, 1, 2],
        [0, 0, 1, 1, 2],
        [0, 0, 0, 0, 3],
        [0, 1, 1, 1, 1],
        [0, 0, 0, 0, 3],
        [0, 1, 1, 1, 1],
        [1, 1, 1, 1, 1],
        [0, 0, 0, 1, 3],
        [1, 0, 0, 0, 3],
        [0, 1, 0, 1, 1],
        [1, 0, 0, 1, 2],
        [0, 0, 0, 1, 3],
        [1, 1, 1, 1, 1],
        [0, 1, 1, 1, 1],
        [1, 1, 1, 1, 1],
        [0, 1, 1, 1, 1],
        [0, 1, 1, 1, 1],
        [1, 1, 1, 1, 1],
        [1, 0, 0, 1, 3],
        [0, 0, 0, 0, 3],
        [1, 1, 1, 1, 1],
        [1, 0, 0, 1, 2],
        [1, 0, 1, 0, 3],
        [0, 1, 0, 1, 1],
        [0, 0, 0, 1, 3],
        [0, 1, 0, 1, 1],
        [1, 1, 0, 1, 1],
        [0, 1, 1, 1, 1],
        [0, 1, 1, 1, 1],
        [1, 1, 1, 1, 1],
        [1, 0, 0, 1, 3],
        [1, 0, 0, 1, 2],
        [0, 1, 1, 1, 1],
        [1, 0, 1, 1, 3],
        [1, 1, 1, 1, 1],
        [0, 1, 0, 1, 1],
        [1, 1, 0, 1, 1],
        [1, 0, 1, 1, 3],
        [1, 1, 1, 1, 1],
        [0, 0, 0, 0, 3],
        [1, 1, 1, 1, 1],
        [0, 0, 0, 1, 3],
        [0, 0, 0, 1, 3],
        [1, 1, 1, 1, 1],
        [0, 1, 1, 1, 1],
        [0, 0, 0, 1, 2],
        [0, 1, 1, 1, 1]
    ]
)

TEST_DATA_DRUG_EFFECTS_A_OBSERVED_DIFF = 0.24
TEST_DATA_DRUG_EFFECTS_A_ORACLE_DIFF = 0.41
TEST_DATA_DRUG_EFFECTS_B_OBSERVED_DIFF = -0.40
TEST_DATA_DRUG_EFFECTS_B_ORACLE_DIFF = -0.06
TEST_DATA_DRUG_EFFECTS_DELTA = 0.15
TEST_DATA_DRUG_EFFECTS_FOLLOWUP_LAST_TIMESTEP = 2
TEST_DATA_PERSON_TIME_LAST_STEP_HAZARD_0 = 0.38
TEST_DATA_PERSON_TIME_LAST_STEP_HAZARD_1 = 0.28
TEST_DATA_PERSON_TIME_LAST_STEP_SURVIVAL_0 = 0.33
TEST_DATA_PERSON_TIME_LAST_STEP_SURVIVAL_1 = 0.58

"""
Test Data for Person-Time Format Conversion
-------------------------------------------
"""
TEST_DATA_PERSON_TIME_INPUT = pd.DataFrame(
    columns=['id', 'age', 'height', 'a', 'y', 't'],
    data=[
        [1, 22, 170, 0, 1, 2],
        [2, 40, 180, 1, 0, 1],
        [3, 30, 165, 1, 0, 2]
    ],
).set_index('id')

TEST_DATA_PERSON_TIME_OUTPUT = pd.DataFrame(
    columns=['id', 'age', 'height', 'a', 'y', 't'],
    data=[
        [1, 22, 170, 0, 0, 0],
        [1, 22, 170, 0, 0, 1],
        [1, 22, 170, 0, 1, 2],
        [2, 40, 180, 1, 0, 0],
        [2, 40, 180, 1, 0, 1],
        [3, 30, 165, 1, 0, 0],
        [3, 30, 165, 1, 0, 1],
        [3, 30, 165, 1, 0, 2]
    ],
).set_index('id')


def fit_synthetic_data(model_cls, params, test_data):
    """
    Test util that runs the following steps:
    1. Initializes a causal survival model.
    2. Fits simulated test data.
    3. Estimates adjusted diff in survival curves between two treatment groups.
    Args:
        model_cls:  causal survival class reference
        params: constructor params
        test_data: DataFrame with covariates 'x_0', 'x_1', treatment assignment 'a', outcome 'y' and time to event 't'

    Returns:
        Adjusted diff in survival curves at last time step
    """
    model = model_cls(**params)
    X = test_data[['x_0', 'x_1']]
    a = test_data['a']
    y = test_data['y']
    t = test_data['t']
    model.fit(X=X, a=a, t=t, y=y)

    # Generate survival curves
    adjusted_curves = model.estimate_population_outcome(X=X, a=a, t=t, y=y)
    adjusted_diff = adjusted_curves[1][TEST_DATA_DRUG_EFFECTS_FOLLOWUP_LAST_TIMESTEP] - adjusted_curves[0][
        TEST_DATA_DRUG_EFFECTS_FOLLOWUP_LAST_TIMESTEP]

    return adjusted_diff


class TestUtils(unittest.TestCase):
    """"
    Test utility methods for causal survival analysis (data handling, etc.)
    """
    def setUp(self) -> None:
        np.random.seed(RANDOM_SEED)

    def test_person_time_expansion(self):
        res = get_person_time_df(a=TEST_DATA_PERSON_TIME_INPUT['a'],
                                 t=TEST_DATA_PERSON_TIME_INPUT['t'],
                                 y=TEST_DATA_PERSON_TIME_INPUT['y'],
                                 X=TEST_DATA_PERSON_TIME_INPUT[['age', 'height']])
        pd.testing.assert_frame_equal(res, TEST_DATA_PERSON_TIME_OUTPUT, check_dtype=False)

    def test_safe_join(self):
        X = TEST_DATA_PERSON_TIME_INPUT.copy()
        uniquely_named_series = [pd.Series(data=0, name='new_col1', index=X.index),
                                 pd.Series(data=0, name='new_col2', index=X.index)]
        duplicate_named_series = [pd.Series(data=0, name='new_col', index=X.index),
                                  pd.Series(data=0, name='new_col', index=X.index)]
        duplicate_in_df_named_series = [pd.Series(data=0, name='age', index=X.index),
                                        pd.Series(data=0, name='height', index=X.index)]

        # Unique Series name - should not rename
        # =======================================
        # Series only
        res, new_names = safe_join(df=None, list_of_series=uniquely_named_series, return_series_names=True)
        self.assertEqual(list(res.columns), [s.name for s in uniquely_named_series])  # res is concat of Series
        self.assertEqual(new_names, [s.name for s in uniquely_named_series])  # Series names UNCHANGED
        self.assertEqual(res.shape, (X.shape[0], len(uniquely_named_series)))

        # DataFrame + Series
        res, new_names = safe_join(df=X, list_of_series=uniquely_named_series, return_series_names=True)
        self.assertEqual(list(res.columns)[:X.shape[1]], list(X.columns))  # DataFrame col names UNCHANGED
        self.assertEqual(new_names, [s.name for s in uniquely_named_series])  # Series names UNCHANGED
        self.assertEqual(res.shape, (X.shape[0], X.shape[1] + len(uniquely_named_series)))

        # Duplicate Series name - should add random suffix
        # ================================================
        # Series only
        res, new_names = safe_join(df=None, list_of_series=duplicate_named_series, return_series_names=True)
        self.assertNotEqual(new_names, [s.name for s in duplicate_named_series])  # Series names CHANGED
        self.assertEqual(res.shape, (X.shape[0], len(duplicate_named_series)))

        # DataFrame + Series
        res, new_names = safe_join(df=X, list_of_series=duplicate_named_series, return_series_names=True)
        self.assertEqual(list(res.columns)[:X.shape[1]], list(X.columns))  # DataFrame col names UNCHANGED
        self.assertNotEqual(new_names, [s.name for s in duplicate_named_series])  # Series names CHANGED
        self.assertEqual(res.shape, (X.shape[0], X.shape[1] + len(duplicate_named_series)))

        # DataFrame + Series with names already in DataFrame
        res, new_names = safe_join(df=X, list_of_series=duplicate_in_df_named_series, return_series_names=True)
        self.assertEqual(list(res.columns)[:X.shape[1]], list(X.columns))  # DataFrame col names UNCHANGED
        self.assertNotEqual(new_names, [s.name for s in duplicate_in_df_named_series])
        self.assertEqual(res.shape, (X.shape[0], X.shape[1] + len(duplicate_in_df_named_series)))


class TestNHEFS(unittest.TestCase):
    """
    Test multiple models on the NHEFS dataset.
    Details: https://www.hsph.harvard.edu/miguel-hernan/causal-inference-book/
    """
    def setUp(self) -> None:
        np.random.seed(RANDOM_SEED)
        data = load_nhefs_survival()
        self.X, self.a, self.t, self.y = data.X, data.a, data.t, data.y

        # Init various multiple models
        self.estimators = {
            'observed_non_parametric': MarginalSurvival(),
            'observed_parametric': MarginalSurvival(survival_model=LogisticRegression(max_iter=2000)),
            'ipw_non_parametric': WeightedSurvival(
                weight_model=IPW(LogisticRegression(max_iter=4000), use_stabilized=True),
                survival_model=None),
            'ipw_parametric': WeightedSurvival(weight_model=IPW(LogisticRegression(max_iter=4000), use_stabilized=True),
                                               survival_model=LogisticRegression(max_iter=4000)),
            'ipw_parametric_pipeline': WeightedSurvival(
                weight_model=IPW(LogisticRegression(max_iter=4000), use_stabilized=True),
                survival_model=Pipeline(
                    [('transform', PolynomialFeatures(degree=2)), ('LR', LogisticRegression(max_iter=1000, C=2))])),
            'standardization_non_stratified': StandardizedSurvival(survival_model=LogisticRegression(max_iter=4000),
                                                                   stratify=False),
            'standardization_stratified': StandardizedSurvival(survival_model=LogisticRegression(max_iter=4000),
                                                               stratify=True),
        }

    def test_nhefs(self, plot=False):
        EXPECTED_SURVIVAL_DIFFS = {'observed_non_parametric': 6,
                                   'observed_parametric': 6,
                                   'ipw_non_parametric': 1,
                                   'ipw_parametric': 1,
                                   'ipw_parametric_pipeline': 1,
                                   'standardization_non_stratified': 1,
                                   'standardization_stratified': 1,
                                   }
        TEST_DELTA = 2

        res = {}
        for estimator_name, estimator in self.estimators.items():
            estimator.fit(X=self.X, a=self.a, t=self.t, y=self.y)
            survival_curves = estimator.estimate_population_outcome(X=self.X, a=self.a, t=self.t, y=self.y)
            res[estimator_name] = survival_curves

        for estimator_name, survival_curve in res.items():
            surv_non_qsmk = 100.0 * float(survival_curve[0].iloc[-1])
            surv_qsmk = 100.0 * float(survival_curve[1].iloc[-1])
            self.assertAlmostEqual(surv_non_qsmk - surv_qsmk, EXPECTED_SURVIVAL_DIFFS[estimator_name], delta=TEST_DELTA)

        # Plots - for debugging purposes
        if plot:
            import matplotlib.pyplot as plt
            import itertools

            grid_dims = (int(np.ceil(np.sqrt(len(self.estimators)))), int(np.round(np.sqrt(len(self.estimators)))))
            grid_indices = itertools.product(range(grid_dims[0]), range(grid_dims[1]))
            fig, ax = plt.subplots(*grid_dims)

            estimator_names = list(self.estimators.keys())
            for estimator_name, plot_idx in zip(estimator_names, grid_indices):
                curve = res[estimator_name]
                ax[plot_idx].plot(curve[0])
                ax[plot_idx].plot(curve[1])
                ax[plot_idx].set_title(estimator_name)
                ax[plot_idx].set_ylim(0.5, 1.02)
                ax[plot_idx].grid()

            plt.show()


class TestMarginalOutcome(unittest.TestCase):
    """
    Test marginal (observed) survival curves.
    """
    def setUp(self) -> None:
        np.random.seed(RANDOM_SEED)

    def test_marginal_outcome_non_parametric_beneficial_effect(self):
        test_data = TEST_DATA_TTE_DRUG_EFFECTS['A']
        model = MarginalSurvival
        params = {}
        observed_diff = fit_synthetic_data(model_cls=model, params=params, test_data=test_data)

        self.assertAlmostEqual(observed_diff, TEST_DATA_DRUG_EFFECTS_A_OBSERVED_DIFF,
                               delta=TEST_DATA_DRUG_EFFECTS_DELTA)

    def test_marginal_outcome_parametric_null_effect(self):
        test_data = TEST_DATA_TTE_DRUG_EFFECTS['B']
        model = MarginalSurvival
        params = {'survival_model': LogisticRegression(max_iter=4000, C=10)}
        observed_diff = fit_synthetic_data(model_cls=model, params=params, test_data=test_data)

        self.assertAlmostEqual(observed_diff, TEST_DATA_DRUG_EFFECTS_B_OBSERVED_DIFF,
                               delta=TEST_DATA_DRUG_EFFECTS_DELTA)


class TestIPW(unittest.TestCase):
    """
    Test inverse propensity weighted survival curves.
    """
    def setUp(self) -> None:
        np.random.seed(RANDOM_SEED)

    def test_ipw_beneficial_effect(self):
        test_data = TEST_DATA_TTE_DRUG_EFFECTS['A']
        model = WeightedSurvival
        params = {'weight_model': IPW(LogisticRegression(max_iter=10000), use_stabilized=True)}
        adjusted_diff = fit_synthetic_data(model_cls=model, params=params, test_data=test_data)

        self.assertAlmostEqual(adjusted_diff, TEST_DATA_DRUG_EFFECTS_A_ORACLE_DIFF,
                               delta=TEST_DATA_DRUG_EFFECTS_DELTA)

    def test_ipw_null_effect(self):
        test_data = TEST_DATA_TTE_DRUG_EFFECTS['B']
        model = WeightedSurvival
        params = {'weight_model': IPW(LogisticRegression(max_iter=10000), use_stabilized=True)}
        adjusted_diff = fit_synthetic_data(model_cls=model, params=params, test_data=test_data)

        self.assertAlmostEqual(adjusted_diff, TEST_DATA_DRUG_EFFECTS_B_ORACLE_DIFF,
                               delta=TEST_DATA_DRUG_EFFECTS_DELTA)

    def test_unnamed_input(self):
        test_data = TEST_DATA_TTE_DRUG_EFFECTS['A']
        X = test_data[['x_0', 'x_1']]
        a = test_data['a']
        y = test_data['y']
        t = test_data['t']

        X.index.name = None
        a.name = None
        y.name = None
        t.name = None
        ipw = WeightedSurvival(weight_model=IPW(LogisticRegression()), survival_model=LogisticRegression())
        ipw.fit(X, a, t, y)
        outcomes = ipw.estimate_population_outcome(X=X, a=a, t=t, y=y)

        self.assertEqual(outcomes.index.name, "t")  # Default time name when canonizing names
        self.assertEqual(outcomes.columns.name, "a")  # Default time name when canonizing names


class TestStandardization(unittest.TestCase):
    """
    Test standardized survival curves.
    """
    def setUp(self) -> None:
        np.random.seed(RANDOM_SEED)

    def test_standardization_non_stratified_beneficial_effect(self):
        test_data = TEST_DATA_TTE_DRUG_EFFECTS['A']
        model = StandardizedSurvival
        params = {'survival_model': LogisticRegression(max_iter=2000, C=10), 'stratify': False}
        adjusted_diff = fit_synthetic_data(model_cls=model, params=params, test_data=test_data)

        self.assertAlmostEqual(adjusted_diff, TEST_DATA_DRUG_EFFECTS_A_ORACLE_DIFF,
                               delta=TEST_DATA_DRUG_EFFECTS_DELTA)

    def test_standardization_stratified_beneficial_effect(self):
        test_data = TEST_DATA_TTE_DRUG_EFFECTS['A']
        model = StandardizedSurvival
        params = {'survival_model': LogisticRegression(max_iter=2000, C=10), 'stratify': True}
        adjusted_diff = fit_synthetic_data(model_cls=model, params=params, test_data=test_data)

        self.assertAlmostEqual(adjusted_diff, TEST_DATA_DRUG_EFFECTS_A_ORACLE_DIFF,
                               delta=TEST_DATA_DRUG_EFFECTS_DELTA)

    def test_standardization_non_stratified_null_effect(self):
        test_data = TEST_DATA_TTE_DRUG_EFFECTS['B']
        model = StandardizedSurvival
        params = {'survival_model': LogisticRegression(max_iter=2000), 'stratify': False}
        adjusted_diff = fit_synthetic_data(model_cls=model, params=params, test_data=test_data)

        self.assertAlmostEqual(adjusted_diff, TEST_DATA_DRUG_EFFECTS_B_ORACLE_DIFF,
                               delta=TEST_DATA_DRUG_EFFECTS_DELTA)

    def test_standardization_stratified_null_effect(self):
        test_data = TEST_DATA_TTE_DRUG_EFFECTS['B']
        model = StandardizedSurvival
        params = {'survival_model': LogisticRegression(max_iter=2000), 'stratify': True}
        adjusted_diff = fit_synthetic_data(model_cls=model, params=params, test_data=test_data)

        self.assertAlmostEqual(adjusted_diff, TEST_DATA_DRUG_EFFECTS_B_ORACLE_DIFF,
                               delta=TEST_DATA_DRUG_EFFECTS_DELTA)

    def test_weighted_standardization_stratified(self):
        test_data = TEST_DATA_TTE_DRUG_EFFECTS['A']
        model = WeightedStandardizedSurvival
        params = {'weight_model': IPW(LogisticRegression(max_iter=1000), use_stabilized=True),
                  'survival_model': LogisticRegression(max_iter=1000),
                  'stratify': True}
        adjusted_diff = fit_synthetic_data(model_cls=model, params=params, test_data=test_data)

        self.assertAlmostEqual(adjusted_diff, TEST_DATA_DRUG_EFFECTS_A_ORACLE_DIFF,
                               delta=TEST_DATA_DRUG_EFFECTS_DELTA)

    def test_weighted_standardization_non_stratified(self):
        test_data = TEST_DATA_TTE_DRUG_EFFECTS['A']
        model = WeightedStandardizedSurvival
        params = {'weight_model': IPW(LogisticRegression(max_iter=2000), use_stabilized=True),
                  'survival_model': LogisticRegression(max_iter=4000, C=5),
                  'stratify': False}
        adjusted_diff = fit_synthetic_data(model_cls=model, params=params, test_data=test_data)

        self.assertAlmostEqual(adjusted_diff, TEST_DATA_DRUG_EFFECTS_A_ORACLE_DIFF,
                               delta=TEST_DATA_DRUG_EFFECTS_DELTA)

    def test_unnamed_input(self):
        test_data = TEST_DATA_TTE_DRUG_EFFECTS['A']
        X = test_data[['x_0', 'x_1']]
        a = test_data['a']
        y = test_data['y']
        t = test_data['t']

        X.index.name = None
        a.name = None
        y.name = None
        t.name = None
        std = StandardizedSurvival(survival_model=LogisticRegression())
        std.fit(X, a, t, y)
        outcomes = std.estimate_population_outcome(X=X, a=a, t=t, y=y)

        self.assertEqual(outcomes.index.name, "t")  # Default time name when canonizing names
        self.assertEqual(outcomes.columns.name, "a")  # Default time name when canonizing names


@unittest.skipUnless(LIFELINES_FOUND, 'lifelines not found')
class TestLifelines(unittest.TestCase):
    """
    Test integration with 'lifelines' Python package.
    """
    def setUp(self) -> None:
        test_data = TEST_DATA_TTE_DRUG_EFFECTS['A']
        self.X = test_data[['x_0', 'x_1']]
        self.a = test_data['a']
        self.y = test_data['y']
        self.t = test_data['t']

    def test_weighted_kaplan_meier_curves(self):
        weighted_survival = WeightedSurvival(
            weight_model=IPW(LogisticRegression(max_iter=10000, C=10), use_stabilized=True),
            survival_model=None
        )
        weighted_survival.fit(self.X, self.a)
        curves_causallib = weighted_survival.estimate_population_outcome(self.X, self.a, self.t, self.y)

        weighted_survival_lifelines_km = WeightedSurvival(
            weight_model=IPW(LogisticRegression(max_iter=10000, C=10), use_stabilized=True),
            survival_model=lifelines.KaplanMeierFitter()
        )
        weighted_survival_lifelines_km.fit(self.X, self.a)
        curves_causallib_lifelines = weighted_survival_lifelines_km.estimate_population_outcome(self.X, self.a, self.t, self.y)

        np.testing.assert_array_almost_equal(curves_causallib, curves_causallib_lifelines, decimal=8)

    def test_marginal_kaplan_meier_curves(self):
        marginal_survival = MarginalSurvival(survival_model=None)
        marginal_survival.fit(self.X, self.a)
        marginal_curves_causallib = marginal_survival.estimate_population_outcome(self.X, self.a, self.t, self.y)

        marginal_survival_lifelines = MarginalSurvival(survival_model=lifelines.KaplanMeierFitter())
        marginal_survival_lifelines.fit(self.X, self.a)
        marginal_curves_causallib_lifelines = marginal_survival_lifelines.estimate_population_outcome(self.X, self.a, self.t, self.y)

        lifelines_km_a0 = lifelines.KaplanMeierFitter()
        lifelines_km_a0.fit(durations=self.t[self.a == 0], event_observed=self.y[self.a == 0])
        lifelines_km_a1 = lifelines.KaplanMeierFitter()
        lifelines_km_a1.fit(durations=self.t[self.a == 1], event_observed=self.y[self.a == 1])
        marginal_curves_lifelines = pd.DataFrame({0: lifelines_km_a0.predict(sorted(self.t.unique())),
                                                  1: lifelines_km_a1.predict(sorted(self.t.unique()))})
        marginal_curves_lifelines.columns.name = 'a'
        marginal_curves_lifelines.index.name = 't'

        pd.testing.assert_frame_equal(marginal_curves_causallib, marginal_curves_causallib_lifelines)
        pd.testing.assert_frame_equal(marginal_curves_causallib, marginal_curves_lifelines)

    def test_cox(self):
        standardized_survival_cox = StandardizedSurvival(survival_model=lifelines.CoxPHFitter())
        standardized_survival_cox.fit(self.X, self.a, self.t, self.y)
        _ = standardized_survival_cox.estimate_population_outcome(self.X, self.a, self.t, self.y)
        # not validating results - only testing for pass/fail

    def test_fit_kwargs(self):
        ipw = IPW(learner=LogisticRegression(max_iter=1000))
        weighted_standardized_survival = WeightedStandardizedSurvival(survival_model=lifelines.CoxPHFitter(),
                                                                      weight_model=ipw)

        # Without fit_kwargs - should raise StatisticalWarning with a suggestion to pass robust=True in fit
        with self.assertWarns(lifelines.exceptions.StatisticalWarning):
            weighted_standardized_survival.fit(self.X, self.a, self.t, self.y)

        # With fit_kwargs - should not raise StatisticalWarning (might raise other warnings, though)
        with self.assertRaises(AssertionError):  # negation workaround since there's no assertNotWarns
            with self.assertWarns(lifelines.exceptions.StatisticalWarning):
                weighted_standardized_survival.fit(self.X, self.a, self.t, self.y, fit_kwargs={'robust': True})


class TestFeatureTransform(unittest.TestCase):
    """
    Test scikit-learn Pipeline/transform as a base learner for standardized survival curves.
    """
    def setUp(self) -> None:
        test_data = TEST_DATA_TTE_DRUG_EFFECTS['A']
        self.X = test_data[['x_0', 'x_1']]
        self.a = test_data['a']
        self.y = test_data['y']
        self.t = test_data['t']
        self.eps = 0.2

        feature_transform = PolynomialFeatures(degree=2)
        learner = LogisticRegression(max_iter=4000, C=0.8)
        self.pipeline = Pipeline([('transform', feature_transform), ('LR', learner)])

    def test_regression_with_transform(self):
        standardized_survival = StandardizedSurvival(survival_model=LogisticRegression(max_iter=1000), stratify=False)
        standardized_survival.fit(self.X, self.a, self.t, self.y)
        curves = standardized_survival.estimate_population_outcome(self.X, self.a, self.t, self.y)

        standardized_survival_poly = StandardizedSurvival(survival_model=self.pipeline, stratify=False)
        standardized_survival_poly.fit(self.X, self.a, self.t, self.y)
        curves_poly = standardized_survival_poly.estimate_population_outcome(self.X, self.a, self.t, self.y)

        diff = np.abs(curves - curves_poly)
        condition = (diff < self.eps).all(axis=None)
        self.assertTrue(condition)



"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on April 4, 2021
"""
import unittest
import numpy as np
import pandas as pd
from sklearn.linear_model import LogisticRegression, LinearRegression
from sklearn.ensemble import RandomForestRegressor, RandomForestClassifier
from causallib.estimation.rlearner import RLearner


class TestRlearner(unittest.TestCase):
    @staticmethod
    def create_uninformative_tx_dataset():
        n = 100
        beta = 0.4
        X = pd.Series(np.random.normal(size=n))
        a = pd.Series([0] * (n // 2) + [1] * (n // 2))
        y = X.mul(beta)
        return {"X": X.to_frame(), "a": a, "y": y, "beta": beta}

    @staticmethod
    def create_uninformative_ox_dataset():
        n = 100
        beta = 0.4
        X = pd.DataFrame(np.random.normal(size=(n, 5)))
        a = pd.Series([0] * (n // 2) + [1] * (n // 2))
        y = a.mul(beta)
        return {"X": X, "a": a, "y": y, "beta": beta}

    @staticmethod
    def create_complex_dataset_nie_wagner():
        """
        following setup A in the paper of
          Nie, X., & Wager, S.(2017).
          Quasi - oracle estimation of heterogeneous treatment effects

        """
        from sklearn.preprocessing import PolynomialFeatures

        n = 500
        d = 6
        sigma = 0.5
        X = pd.DataFrame(np.random.uniform(size=(n, d)))
        propensity = np.clip(np.sin(np.pi * X.iloc[:, 0] * X.iloc[:, 1]), 0.1,
                             0.9)
        a = pd.Series(np.random.binomial(1, propensity, n))
        baseline_outcome = (
                np.sin(np.pi * X.iloc[:, 0] * X.iloc[:, 1])
                + 2 * (X.iloc[:, 2] - 0.5) ** 2
                + X.iloc[:, 3]
                + 0.5 * X.iloc[:, 4]
        )
        tau = (X.iloc[:, 0] + X.iloc[:, 1]) / 2
        noise = pd.Series(np.random.normal(size=(n,)))
        y = baseline_outcome + (a - 0.5) * tau + sigma * noise

        # pairwise interactions with 7 degrees of freedom
        poly = PolynomialFeatures(7, interaction_only=True, include_bias=False)
        X_pn = pd.DataFrame(poly.fit_transform(X))
        return {"X": X, "X_pn": X_pn, "a": a, "y": y, "tau": tau}

    @staticmethod
    def create_complex_data_for_ate_victor():
        """
        following the simulation at:
        https://www.r-bloggers.com/2017/06/cross-fitting-double-machine-learning-estimator/
        http://aeturrell.com/2018/02/10/econometrics-in-python-partI-ML/

        y = a * \tau + G(x,b) + u
        a = M(x,b) + v

        where u, v are noise terms and M() and G() are function of the covariates
        for the treatment assignment and the outcome surface, respectively.
        """
        from sklearn.datasets import make_spd_matrix

        def g(x):
            return np.power(np.sin(x), 2)

        def m(x, nu=0.0, gamma=1.0):
            return 0.5 / np.pi * (np.sinh(gamma)) / (
                    np.cosh(gamma) - np.cos(x - nu))

        n = 500
        tau = 0.5
        k = 10
        b = 1 / np.arange(1, k + 1)
        sigma = make_spd_matrix(k, random_state=1234)

        X = pd.DataFrame(
            np.random.multivariate_normal(np.ones(k), sigma, size=(n,))
        )
        G = g(np.dot(X, b))
        M = m(np.dot(X, b))
        a = pd.Series(
            M + np.random.standard_normal(size=(n,))
        )
        y = pd.Series(
            np.dot(tau, a) + G + np.random.standard_normal(size=(n,))
        )

        return {"X": X, "a": a, "y": y, "tau": tau}

    def ensure_prepare_data(self, estimator):
        data = self.create_uninformative_ox_dataset()

        with self.subTest(
                "Test that the dimension of the feature space is "
                "equal to the input space"
        ):
            X_outcome, X_treatment, X_effect = estimator._prepare_data(
                data["X"])
            self.assertListEqual(
                [data["X"].shape[1]] * 2,
                [X_outcome.shape[1], X_treatment.shape[1]]
            )
            self.assertEqual(data["X"].shape[1] + 1, X_effect.shape[1])

        with self.subTest("Test smaller feature space for the outcome model"):
            estimator.outcome_covariates = [0, 1, 2]
            X_outcome, X_treatment, X_effect = estimator._prepare_data(
                data["X"])
            self.assertEqual(data["X"].shape[1], X_treatment.shape[1])
            self.assertEqual(3, X_outcome.shape[1])

        with self.subTest("Test no covariate in the effect model"):
            estimator.effect_covariates = list()
            X_outcome, X_treatment, X_effect = estimator._prepare_data(
                data["X"])
            self.assertEqual(1, X_effect.shape[1])

    def ensure_nuisance_models_are_fitted(self, estimator):
        data = self.create_uninformative_ox_dataset()
        estimator.fit(data["X"], data["a"], data["y"])
        self.assertTrue(hasattr(estimator.outcome_model, "coef_"))
        self.assertTrue(hasattr(estimator.treatment_model, "coef_"))

    def ensure_homogeneous_effect_for_none_X_effect(self, estimator):
        """
        when the effect_covariates is equal to list(), we require same effect
        across different covariates.
        """
        data = self.create_uninformative_tx_dataset()
        estimator = estimator.__class__(
            outcome_model=estimator.outcome_model,
            treatment_model=estimator.treatment_model,
            effect_model=estimator.effect_model,
            effect_covariates=list(),
        )

        estimator.fit(data["X"], data["a"], data["y"])
        estimated_effect = estimator.estimate_individual_effect(data["X"])
        np.testing.assert_array_equal(  # Array is constant
            estimated_effect[0], estimated_effect,
        )

    def ensure_estimate_effect_continuous_treatment(self, estimator):
        data = self.create_complex_data_for_ate_victor()
        estimator = estimator.__class__(
            outcome_model=estimator.outcome_model,
            treatment_model=LinearRegression(normalize=True),
            effect_model=estimator.effect_model,
        )
        with self.subTest("Test fit"):
            estimator.fit(data["X"], data["a"], data["y"])
            self.assertTrue(True)  # Dummy assert, didn't crash
        with self.subTest("Check prediction"):
            individual_cf_outcome = estimator.estimate_individual_outcome(
                data["X"], data["a"], treatment_values=[-0.2, 0.32, 1.1]
            )
            self.assertEqual(3, individual_cf_outcome.shape[1])

    def ensure_naive_model_for_uninformative_nuisance_models_rlearnerlinear(
            self, estimator
    ):
        """
        without any ability to predict the response surface and the treatment
        assignment, the rlearner narrows to a simple naive estimator
        """
        data = self.create_uninformative_ox_dataset()
        X_effect = estimator._extract_effect_model_data(data["X"])

        estimator._fit_linear_effect_model(X_effect, res_a=data["a"],
                                           res_y=data["y"])
        estimated_effect = estimator.estimate_individual_effect(
            data["X"]).mean()
        self.assertAlmostEqual(estimated_effect, data["beta"])
        np.testing.assert_array_almost_equal(
            estimator.effect_model.coef_[1:], [0] * data["X"].shape[1]
        )

    def ensure_cross_fitting_is_not_overfitting(self, estimator):
        """
        In cross-fitting the prediction are done on held-out data and
        """
        from sklearn.tree import DecisionTreeClassifier, DecisionTreeRegressor
        from sklearn.metrics import mean_squared_error, roc_auc_score

        decision_tree_rg = DecisionTreeRegressor(max_depth=12)
        decision_tree_clf = DecisionTreeClassifier(max_depth=12)

        data = self.create_complex_dataset_nie_wagner()
        # fit and predict on the same data
        decision_tree_rg.fit(data["X"], data["y"])
        decision_tree_clf.fit(data["X"], data["a"])
        pred_y_decision_tree = decision_tree_rg.predict(data["X"])
        pred_a_decision_tree = decision_tree_clf.predict_proba(data["X"])[:, 1]

        estimator = estimator.__class__(
            outcome_model=decision_tree_rg,
            treatment_model=decision_tree_clf,
            effect_model=estimator.effect_model,
        )
        X_outcome, X_treatment, X_effect = estimator._prepare_data(data["X"])
        # cross-fitting
        pred_y, _ = estimator._fit_and_predict_model(
            estimator.outcome_model, X_outcome, data["y"], predict_proba=False
        )
        pred_a, _ = estimator._fit_and_predict_model(
            estimator.treatment_model, X_treatment, data["a"],
            predict_proba=True
        )
        self.assertLess(
            mean_squared_error(data["y"], pred_y_decision_tree),
            mean_squared_error(data["y"], pred_y)
        )

        self.assertLess(
            roc_auc_score(data["a"], np.round(pred_a)),
            roc_auc_score(data["a"], np.round(pred_a_decision_tree))
        )

    def ensure_enforce_fit_intercept_false(self, estimator):
        data = self.create_uninformative_ox_dataset()

        with self.assertWarns(UserWarning):
            # Trigger a warning.
            estimator_fit_intercept = estimator.__class__(
                outcome_model=estimator.outcome_model,
                treatment_model=estimator.treatment_model,
                effect_model=LinearRegression(fit_intercept=True),
            )

            estimator_fit_intercept.fit(data["X"], data["a"], data["y"])
            self.assertFalse(estimator_fit_intercept.effect_model.fit_intercept)

            estimator.fit(data["X"], data["a"], data["y"])
            self.assertAlmostEqual(
                estimator.effect_model.coef_[0],
                estimator_fit_intercept.effect_model.coef_[0],
                places=3,
            )

    def ensure_estimate_ate_as_in_literature(self, estimator):
        from sklearn.ensemble import RandomForestRegressor

        estimator = estimator.__class__(
            outcome_model=RandomForestRegressor(max_depth=2),
            treatment_model=RandomForestRegressor(max_depth=2),
            effect_model=LinearRegression(fit_intercept=False),
            effect_covariates=list(),
        )

        res = []
        for i in range(100):
            data = self.create_complex_data_for_ate_victor()
            estimator.fit(data["X"], data["a"], data["y"])
            res.append(np.mean(estimator.estimate_individual_effect(data["X"])))
        pred_ate = np.mean(res)
        self.assertAlmostEqual(pred_ate, 0.5, places=1)

    def ensure_pipeline_linear_gridsearch(self, estimator):
        from sklearn.linear_model import LassoCV
        from sklearn.model_selection import GridSearchCV

        data = self.create_complex_dataset_nie_wagner()

        parameters = {"max_depth": [3, 7, 10]}
        gs = GridSearchCV(RandomForestClassifier(), parameters)
        with self.subTest("Test initialization with models and gridsearch"):
            estimator = estimator.__class__(
                outcome_model=LassoCV(normalize=True),
                treatment_model=gs,
                effect_model=LassoCV(fit_intercept=False, normalize=True),
                n_splits=2,
            )
            self.assertTrue(True)  # Dummy assert for not thrown exception

        with self.subTest("Test fit with pipeline estimator"):
            estimator.fit(data["X"], data["a"], data["y"])
            self.assertTrue(True)  # Dummy assert for not thrown exception

        with self.subTest("Test 'predict' with pipeline learner"):
            estimator.estimate_individual_outcome(data["X"], data["a"])
            self.assertTrue(True)  # Dummy assert for not thrown exception

    def ensure_warning_when_linear_model_and_non_parametric_is_true(self):
        estimator = RLearner(
            outcome_model=LinearRegression(),
            treatment_model=LogisticRegression(),
            effect_model=LinearRegression(fit_intercept=False),
            non_parametric=True,
        )
        data = self.create_uninformative_ox_dataset()
        with self.assertWarns(UserWarning):
            estimator.fit(data["X"], data["a"], data["y"])

    def ensure_rlearner_can_be_evaluated(self):
        """ensure that Rlearner can be evaluated"""
        from causallib.evaluation import evaluate
        from causallib.evaluation.results import ContinuousOutcomeEvaluationResults
        data = self.create_complex_dataset_nie_wagner()
        self.estimator.fit(data["X"], data["a"], data["y"])
        with self.subTest("Test evaluate R-learner"):
            evaluation_results =evaluate(self.estimator, data['X'], data['a'], data['y'])
            self.assertIsNotNone(evaluation_results)  # Dummy assert for not thrown exception
            self.assertIsInstance(evaluation_results, ContinuousOutcomeEvaluationResults)

class TestRLearnerLinear(TestRlearner):
    @classmethod
    def setUpClass(cls):
        TestRlearner.setUpClass()
        # Avoids regularization of the model:
        treatment_model = LogisticRegression(solver="sag", penalty="none")
        outcome_model = LinearRegression(normalize=True)
        effect_model = LinearRegression(fit_intercept=False)
        cls.estimator = RLearner(
            outcome_model=outcome_model,
            treatment_model=treatment_model,
            effect_model=effect_model,
            non_parametric=False,
        )

    def test_prepare_data(self):
        self.ensure_prepare_data(self.estimator)

    def test_is_fitted(self):
        self.ensure_nuisance_models_are_fitted(self.estimator)

    def test_homogeneous_effect_for_none_X_effect(self):
        self.ensure_homogeneous_effect_for_none_X_effect(self.estimator)

    def test_cross_fitting_is_not_overfitting(self):
        self.ensure_cross_fitting_is_not_overfitting(self.estimator)

    def test_naive_estimator_for_uninformative_nuisance_models(self):
        self.ensure_naive_model_for_uninformative_nuisance_models_rlearnerlinear(
            self.estimator
        )

    def test_enforce_fit_intercept_false(self):
        self.ensure_enforce_fit_intercept_false(self.estimator)

    def test_estimate_effect_continuous_treatment(self):
        self.ensure_estimate_effect_continuous_treatment(self.estimator)

    def test_pipeline_linear_gridsearch(self):
        self.ensure_pipeline_linear_gridsearch(self.estimator)

    def test_warning_when_linear_model_and_non_parametric_is_true(self):
        self.ensure_warning_when_linear_model_and_non_parametric_is_true()

    def test_rlearner_can_be_evaluated(self):
        self.ensure_rlearner_can_be_evaluated()

    @unittest.skip("long testing procedure")
    def test_compare_literature_ate_results(self):
        self.ensure_estimate_ate_as_in_literature(self.estimator)


class TestRlearnerNonparam(TestRlearner):
    @classmethod
    def setUpClass(cls):
        TestRlearner.setUpClass()
        treatment_model = LogisticRegression(solver="sag", penalty="none")
        outcome_model = LinearRegression(normalize=True)
        effect_model = RandomForestRegressor()
        cls.estimator = RLearner(
            outcome_model=outcome_model,
            treatment_model=treatment_model,
            effect_model=effect_model,
            non_parametric=True,
        )

    def test_prepare_data(self):
        self.ensure_prepare_data(self.estimator)

    def test_homogeneous_effect_for_none_X_effect(self):
        self.ensure_homogeneous_effect_for_none_X_effect(self.estimator)

    def test_is_fitted(self):
        self.ensure_nuisance_models_are_fitted(self.estimator)

    def test_cross_fitting_is_not_overfitting(self):
        self.ensure_cross_fitting_is_not_overfitting(self.estimator)

    def test_estimate_effect_continuous_treatment(self):
        self.ensure_estimate_effect_continuous_treatment(self.estimator)

    @unittest.skip("long testing procedure")
    def test_compare_literature_ate_results(self):
        self.ensure_estimate_ate_as_in_literature(self.estimator)



# (C) Copyright 2020 IBM Corp.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Created on Nov 12, 2020

import unittest
import warnings
import numpy as np
import pandas as pd
from sklearn.datasets import make_classification
from sklearn.linear_model import LogisticRegression
from causallib.estimation import Matching, PropensityMatching
import pickle

from causallib.preprocessing.transformers import PropensityTransformer, MatchingTransformer


class TestMatching(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.n = 100
        cls.k = 10
        cls.random_seed = 5432
        np.random.seed(cls.random_seed)
        cls.data_serial_x = cls._generate_serial_covariates_data(cls.n)
        cls.data_serial_unbalanced_x = cls._generate_serial_covariates_unbalanced_treatment_data(
            cls.n, cls.k)
        cls.data_3feature_linear_effect = cls._generate_3feature_linear_effect_data(
            cls.random_seed)

        cls.test_data = [cls.data_serial_x, cls.data_serial_unbalanced_x,
                         cls.data_3feature_linear_effect]

    @classmethod
    def _generate_serial_covariates_data(cls, n):
        '''
        The simplest dataset. Covariates are 1d integers 0 to n-1. Treatment is
        0 for one set of 0 to n-1, 1 for another set of 0 to n-1. Outcome is 
        random uniform for untreated and 2x random uniform for treated.
        '''
        X = pd.DataFrame(np.hstack([np.arange(n), np.arange(n)]))
        X.columns = [f"x{i}" for i in range(len(X.columns))]
        a = pd.to_numeric(
            pd.Series(np.hstack([np.zeros(n), np.ones(n)]), name="treatment"),
            downcast="integer")

        y = pd.Series(
            np.hstack([np.random.rand(n), 2 * np.random.rand(n)]), name="outcome")
        return X, a, y

    @classmethod
    def _generate_serial_covariates_unbalanced_treatment_data(cls, n, k):
        '''
        The same as above but imbalanced. Covariates are 1d integers 0 to n-1. 
        Treatment is 0 for one set of 0 to n-1, 1 for another set of 0 to k-1. 
        Outcome is equal to covariates for untreated and 2x covariates for treated.
        '''
        X = pd.DataFrame(np.hstack([np.arange(n), np.arange(k)]))
        X.columns = [f"x{i}" for i in range(len(X.columns))]
        a = pd.to_numeric(
            pd.Series(np.hstack([np.zeros(n), np.ones(k)]), name="treatment"),
            downcast="integer")
        y = pd.Series(
            np.hstack([np.arange(n), 2 * np.arange(k)]), name="outcome")

        return X, a, y

    @classmethod
    def _generate_3feature_linear_effect_data(cls, random_state):
        '''
        Two stages of data creation. A generic classification set of 500 samples
        to predict treatment (so that the propensity model should find something).
        Outcome is theta*X + 5*a so there is a linear effect of strength 5 plus a 
        covariate dependency via the vector theta which is hardcoded to (0.1, 0.2, 0.3)
        '''
        X, a = make_classification(
            n_features=3,
            n_samples=500,
            n_informative=2,
            n_redundant=0,
            n_repeated=0,
            n_classes=2,
            n_clusters_per_class=1,
            flip_y=0.0,
            class_sep=1.0,
            random_state=random_state,
        )
        X = (
            pd.DataFrame(X)
            .assign(treatment=a)
            .assign(outcome=X @ (0.1, 0.2, 0.3) + 5 * a)
        )
        a = X.pop("treatment")
        y = X.pop("outcome")
        X.columns = [f"x{i}" for i in range(len(X.columns))]

        return X, a, y

    def setUp(self):
        self.transformer = MatchingTransformer()
        self.matching = Matching()

    def test_that_serial_integer_covariates_match_exactly_with_replacement(self):
        X, a, y = self.data_serial_x

        self.matching.fit(X, a, y)
        self._check_index_matching_for_serial_indices_(X, a)

    def test_that_serial_integer_covariates_match_exactly_with_no_replacement(self):
        X, a, y = self.data_serial_x
        self.matching.with_replacement = False

        self.matching.fit(X, a, y)
        self._check_index_matching_for_serial_indices_(X, a)

    def test_that_serial_integer_covariates_match_exactly_when_unbalanced_with_replacement(
        self,
    ):
        X, a, y = self.data_serial_unbalanced_x

        self.matching.fit(X, a, y)
        self._check_index_matching_for_serial_indices_(X, a)

    def test_that_serial_integer_covariates_match_exactly_when_unbalanced_with_no_replacement(
        self,
    ):
        X, a, y = self.data_serial_unbalanced_x
        self.matching.with_replacement = False

        self.matching.fit(X, a, y)
        self._check_index_matching_for_serial_indices_(X, a)

    def test_transform_for_serial_integer_covariates_when_unbalanced_with_no_replacement(
        self,
    ):
        X, a, y = self.data_serial_unbalanced_x

        self.transformer.set_params(with_replacement=False)
        Xm, am, ym = self.transformer.fit_transform(X, a, y)
        np.testing.assert_array_equal(Xm, X[X <= (X[a == 1].max())].dropna())

    def test_transform_for_serial_integer_covariates_when_unbalanced_with_replacement(
        self,
    ):
        X, a, y = self.data_serial_unbalanced_x
        self.transformer.set_params(with_replacement=True, caliper=0.5)
        Xm, am, ym = self.transformer.fit_transform(X, a, y)
        np.testing.assert_array_equal(Xm, X[X <= (X[a == 1].max())].dropna())

    def test_transform_for_serial_integer_covariates_when_unbalanced_with_replacement_partial_match(
        self,
    ):
        X, a, y = self.data_serial_unbalanced_x
        self.transformer.set_params(
            with_replacement=True, matching_mode="treatment_to_control"
        )
        Xm, am, ym = self.transformer.fit_transform(X, a, y)
        np.testing.assert_array_equal(Xm, X[X <= (X[a == 1].max())].dropna())

        self.transformer.set_params(matching_mode="control_to_treatment")
        Xm, am, ym = self.transformer.fit_transform(X, a, y)

        np.testing.assert_array_equal(Xm, X)

    def test_unsupported_matching_mode_error(self):
        X, a, y = self.data_serial_unbalanced_x
        self.transformer.set_params(matching_mode="i'm not supported")

        with self.assertRaises(NotImplementedError):
            self.transformer.fit_transform(X, a, y)

    def test_effect_of_treatment_on_treated_serial_covariates_when_unbalanced_with_no_replacement(
        self,
    ):
        X, a, y = self.data_serial_unbalanced_x

        self.matching.with_replacement = False
        self.matching.fit(X, a, y)
        Ypop = self.matching.estimate_population_outcome(X, a)
        self.assertAlmostEqual(Ypop[1] / Ypop[0], 2)
        Yind = self.matching.estimate_population_outcome(X, a)
        self.assertAlmostEqual(2 * Yind[0], Yind[1])

    def test_effect_of_treatment_on_treated_serial_covariates_when_unbalanced_with_replacement(
        self,
    ):
        X, a, y = self.data_serial_unbalanced_x

        self.matching.with_replacement = True
        self.matching.caliper = 0.5
        self.matching.fit(X, a, y)

        Ypop = self.matching.estimate_population_outcome(X, a)
        self.assertAlmostEqual(Ypop[1] / Ypop[0], 2)

    def test_propensity_matching_object(self):

        self.matching = PropensityMatching(
            learner=LogisticRegression(solver="liblinear"), with_replacement=True
        )
        X, a, y = self.data_3feature_linear_effect
        # pull naive effect estimate
        naive_ate = y[a == 1].mean() - y[a == 0].mean()

        # now do the no caliper distance and effect estimates
        self.matching.propensity_transform = PropensityTransformer(
            learner=LogisticRegression(solver="liblinear"),
            include_covariates=False
        )
        self.matching.fit(X, a, y)

        self.assertEqual(
            1, self.matching.propensity_transform.transform(X).shape[1])
        self.assertTrue(
            "propensity" in self.matching.propensity_transform.transform(X))
        # verify that values of `a` are higher when propensity is larger
        covariates = (
            self.matching.propensity_transform.transform(X)
            .join(a)
            .sort_values("propensity")
        )
        self.assertLess(
            covariates.iloc[:covariates.shape[0] // 2].treatment.mean(),
            covariates.iloc[covariates.shape[0] // 2:].treatment.mean(),
        )

    def test_match_on_multiple_covariates_with_replacement_and_caliper(self):
        X, a, y = self.data_3feature_linear_effect

        # pull naive effect estimate
        naive_ate = y[a == 1].mean() - y[a == 0].mean()

        # now do the no caliper distance and effect estimates
        self.matching.fit(X, a, y)
        self.matching.with_replacement = True
        self.matching.caliper = None
        self.matching.match(X, a)
        match_distances_no_caliper = np.hstack(
            [self._get_nearest_match_distances_for_treatment_value(t,a) for t in [0, 1]])
        no_caliper_ate = self.matching.estimate_population_outcome(X, a).diff()[
            1]

        # with no caliper we should get exactly as many valid matches as samples
        self.assertEqual(match_distances_no_caliper.shape[0], len(X))
        # and the ate estimate should be closer to the ground truth effect of 5
        self.assertLess(abs(5 - no_caliper_ate), abs(5 - naive_ate))

        self.matching.caliper = np.quantile(match_distances_no_caliper, 0.01)

        # now do the caliper constrained distance and effect estimate
        self.matching.with_replacement = True
        self.matching.match(X, a, use_cached_result=False)
        match_distances_with_caliper = np.hstack(
            [self._get_nearest_match_distances_for_treatment_value(t,a) for t in [0, 1]])
        caliper_ate = self.matching.estimate_population_outcome(X, a).diff()[1]

        # with a caliper there will be fewer matches
        self.assertLess(
            match_distances_with_caliper.shape[0], match_distances_no_caliper.shape[0]
        )
        # check that caliper works by checking the maximal distance is 
        # lower than the caliper
        self.assertLess(match_distances_with_caliper.max(), self.matching.caliper)
        # and the ate estimate should be EVEN closer to the ground truth effect of 5
        self.assertLess(abs(5 - caliper_ate), abs(5 - no_caliper_ate))

    def test_match_on_multiple_covariates_with_no_replacement_and_caliper(self):
        X, a, y = self.data_3feature_linear_effect

        # pull naive effect estimate
        naive_ate = y[a == 1].mean() - y[a == 0].mean()

        # now do the no caliper distance and effect estimates
        self.matching.fit(X, a, y)
        self.matching.with_replacement = False
        self.matching.caliper = None
        self.matching.match(X, a)

        match_distances_no_caliper = np.hstack(
            [self._get_nearest_match_distances_for_treatment_value(t,a) for t in [0, 1]])
        no_caliper_ate = self.matching.estimate_population_outcome(X, a).diff()[
            1]

        # with no caliper we should get exactly as many valid matches as samples
        self.assertEqual(match_distances_no_caliper.shape[0], len(X))
        # and the ate estimate should be exactly as good as the naive estimate because there is no caliper
        # and the groups are balanced
        self.assertAlmostEqual(abs(5 - no_caliper_ate), abs(5 - naive_ate))

        self.matching.caliper = np.quantile(match_distances_no_caliper, 0.01)

        # now do the caliper constrained distance and effect estimate
        self.matching.with_replacement = False
        self.matching.match(X, a, use_cached_result=False)


        match_distances_with_caliper = np.hstack(
            [self._get_nearest_match_distances_for_treatment_value(t,a) for t in [0, 1]])
        caliper_ate = self.matching.estimate_population_outcome(X, a).diff()[1]

        # with a caliper there will be fewer matches
        self.assertLess(
            match_distances_with_caliper.shape[0], match_distances_no_caliper.shape[0]
        )
        # and the ate estimate should be EVEN closer to the ground truth effect of 5
        self.assertLess(abs(5 - caliper_ate), abs(5 - no_caliper_ate))

    def test_match_on_multiple_covariates_with_propensity_added(self):
        X, a, y = self.data_3feature_linear_effect
        # pull naive effect estimate
        naive_ate = y[a == 1].mean() - y[a == 0].mean()

        # now do the no caliper distance and effect estimates
        self.matching.propensity_transform = PropensityTransformer(
            learner=LogisticRegression(solver="liblinear"),
            include_covariates=True
        )
        self.matching.fit(X, a, y)

        self.assertTrue(
            "propensity" in self.matching.propensity_transform.transform(
                X).columns
        )
        self.assertEqual(
            X.shape[1] + 1,
            self.matching.propensity_transform.transform(X).shape[1],
        )
        # verify that values of `a` are higher when propensity is larger
        covariates = (
            self.matching.propensity_transform.transform(X)
            .join(a)
            .sort_values("propensity")
        )
        # after sorting by propensity, there should be more a=1 in the upper
        # half of the list than the lower half, else the propensity model is
        # broken
        self.assertLess(
            covariates.iloc[:covariates.shape[0] // 2].treatment.mean(),
            covariates.iloc[covariates.shape[0] // 2:].treatment.mean(),
        )

        # now do the no caliper distance and effect estimates
        self.matching.caliper = None

        no_caliper_ate = self.matching.estimate_population_outcome(X, a).diff()[
            1]
        # and the ate estimate should be closer to the ground truth effect of 5
        self.assertLess(abs(5 - no_caliper_ate), abs(5 - naive_ate))

    def test_match_on_multiple_covariates_with_propensity_replacing_covariates_for_matching(
        self,
    ):
        X, a, y = self.data_3feature_linear_effect
        # pull naive effect estimate
        naive_ate = y[a == 1].mean() - y[a == 0].mean()

        # now do the no caliper distance and effect estimates
        self.matching.propensity_transform = PropensityTransformer(
            learner=LogisticRegression(solver="liblinear"),
            include_covariates=False
        )
        self.matching.fit(X, a, y)

        self.assertEqual(
            1, self.matching.propensity_transform.transform(X).shape[1])
        self.assertTrue(
            "propensity" in self.matching.propensity_transform.transform(
                X).columns
        )
        # verify that values of `a` are higher when propensity is larger
        covariates = (
            self.matching.propensity_transform.transform(X)
            .join(a)
            .sort_values("propensity")
        )
        # after sorting by propensity, there should be more a=1 in the upper
        # half of the list than the lower half, else the propensity model is
        # broken
        self.assertLess(
            covariates.iloc[:covariates.shape[0] // 2].treatment.mean(),
            covariates.iloc[covariates.shape[0] // 2:].treatment.mean(),
        )

        # now do the caliper distance and effect estimates

        caliper_ate = self.matching.estimate_population_outcome(X, a).diff()[1]
        # and the ate estimate should be closer to the ground truth effect of 5
        self.assertLess(abs(5 - caliper_ate), abs(5 - naive_ate))

    def test_weights_with_no_replacement(self):
        X, a, y = self.data_serial_unbalanced_x

        self.matching.with_replacement = False
        self.matching.fit(X, a, y)
        self.matching.match(X, a)
        weights = self.matching.matches_to_weights()
        np.testing.assert_array_equal(
            weights["control_to_treatment"], weights["treatment_to_control"]
        )

    def test_weights_with_multiple_neighbors(self):
        X, a, y = self.data_serial_unbalanced_x

        self.matching.with_replacement = True
        self.matching.n_neighbors = 1
        self.matching.fit(X, a, y)
        self.matching.match(X, a)
        weights = self.matching.matches_to_weights()
        self.assertEqual(weights.iloc[-1][0], self.n - self.k + 1)
        self.assertEqual(
            weights["treatment_to_control"]
            .iloc[self.k + (self.matching.n_neighbors + 1) // 2: self.n]
            .sum(),
            0,
        )

        self.matching.n_neighbors = 5
        self.matching.match(X, a)
        weights = self.matching.matches_to_weights()
        self.assertEqual(
            weights["treatment_to_control"]
            .iloc[self.k + (self.matching.n_neighbors + 1) // 2: self.n]
            .sum(),
            0,
        )

    def test_exception_for_no_replacement_and_n_neighbors_gt_1(self):
        X, a, y = self.data_serial_x
        self.matching.n_neighbors = 2
        self.matching.with_replacement = False
        self.matching.fit(X, a, y)

        with self.assertRaises(NotImplementedError):
            self.matching.match(X, a)

    def test_exception_for_outcome_estimation_following_failed_matching(self):
        X, a, y = self.data_3feature_linear_effect
        self.matching.caliper = 0
        self.matching.with_replacement = False
        self.matching.fit(X, a, y)
        with self.assertWarns(Warning):
            self.matching.match(X, a)
        with self.assertRaises(ValueError):
            self.matching.estimate_population_outcome(X, a)

    def test_covariate_matching(self):
        from sklearn.exceptions import NotFittedError

        X, a, y = self.data_serial_unbalanced_x
        self.matching.fit(X, a, y)
        self.matching.with_replacement = True
        self.matching.caliper = None

        with self.assertRaises(NotFittedError):
            self.matching.get_covariates_of_matches(0, 1, X)
        self.matching.match(X, a)
        covariates_0_1 = self.matching.get_covariates_of_matches(0, 1, X)
        covariates_1_0 = self.matching.get_covariates_of_matches(1, 0, X)
        self.assertTrue(all(covariates_0_1[["delta"]].values[:, 0] == 0))
        np.testing.assert_array_equal(
            covariates_1_0[["delta"]].values[self.k:, 0],
            np.arange(1, self.n + 1 - self.k),
        )
        np.testing.assert_array_equal(
            covariates_1_0[["delta"]].values[: self.k, 0], np.zeros((self.k,))
        )

    def test_that_out_of_sample_outcome_prediction_finds_fitted_data(self):
        from sklearn.model_selection import train_test_split

        for data in self.test_data:
            X, a, y = data
            Xt, Xv, at, av, yt, yv = train_test_split(
                X, a, y, random_state=self.random_seed)
            self.matching.fit(Xt, at, yt)
            # estimate outcome for validation should pull data from train
            ypredv = self.matching.estimate_individual_outcome(Xv, av, yv)
            ypredv_melt = ypredv.melt(var_name=at.name, value_name=yt.name)
            trainval_match = (
                pd.concat(
                    {"train": pd.concat([yt, at], axis=1),
                     "val": ypredv_melt},
                    sort=True,
                )
                .reset_index()
                .rename(columns={"level_0": "fold", "level_1": "sample_id"})
            )
            # to check that we are finding a training data for every matched validation
            # we create a dataframe combining outcomes for both folds and then
            # group by the predicted outcome and check that every group that has a
            # validation sample also has a training sample, in other words every time
            # a validation sample matches to something, it is matching to a valid
            # sample within the training data
            for gval, gidx in trainval_match.groupby(yt.name).groups.items():
                gx = trainval_match.loc[gidx]
                # if there is a validation with this outcome
                if len(gx[gx.fold == "val"]) > 1:
                    # there must also be a training sample
                    self.assertGreater(len(gx[gx.fold == "train"]), 0)

    def test_classify_task(self):

        X, a, y = self.data_3feature_linear_effect
        y = y.astype(int)
        self.matching.fit(X, a, y)
        self.matching.caliper = None
        ypred = self.matching.estimate_individual_outcome(
            X, a, predict_proba=False)
        # if not all integers will raise errors
        ypred.applymap(str).applymap(int)
        with self.assertRaises(ValueError):
            self.matching.estimate_individual_outcome(
                X, a, predict_proba=True
            ).applymap(str).applymap(int)

    def test_classify_task_with_wrong_inputs_warning_check(self):
        X, a, y = self.data_3feature_linear_effect
        # y = y.astype(int)
        self.matching.fit(X, a, y)
        self.matching.caliper = None
        with self.assertWarns(Warning):
            self.matching.estimate_individual_outcome(
                X, a, predict_proba=False)

    def test_treatment_to_control_control_to_treatment_match(self):
        X, a, y = self.data_serial_unbalanced_x
        self.matching.fit(X, a, y)
        self.matching.estimate_observed_outcome = False

        self.matching.matching_mode = "treatment_to_control"
        ypred = self.matching.estimate_individual_outcome(X, a)
        np.testing.assert_array_equal(ypred[1].values, y[a == 1].values)
        with self.assertRaises(AssertionError):
            np.testing.assert_array_equal(ypred[0].values, y[a == 0].values)

        self.matching.matching_mode = "control_to_treatment"
        ypred = self.matching.estimate_individual_outcome(X, a)
        np.testing.assert_array_equal(ypred[0].values, y[a == 0].values)
        with self.assertRaises(AssertionError):
            np.testing.assert_array_equal(ypred[1].values, y[a == 1].values)

        self.matching.matching_mode = "i'm a mistake"
        with self.assertRaises(NotImplementedError):
            self.matching.estimate_individual_outcome(X, a)

    def _check_index_matching_for_serial_indices_(self, X, a):

        match_df = self.matching.match(X, a)

        matched_indices_1_to_0 = (
            match_df.loc[0][a == 1]
            .matches.apply(lambda x: x[0] if x else np.nan)
            .dropna()
            .values
        )
        matched_indices_0_to_1 = (
            match_df.loc[1][a == 0]
            .matches.apply(lambda x: x[0] if x else np.nan)
            .dropna()
            .values
        )

        zero_indices = X[a == 0].index.values
        one_indices = X[a == 1].index.values

        self.assertEqual(
            sum(np.diff(list(zip(matched_indices_0_to_1, one_indices)))), 0
        )
        self.assertEqual(
            sum(np.diff(list(zip(matched_indices_1_to_0, zero_indices)))), 0
        )

    def test_all_backends_agree(self):
        try:
            import faiss
        except ImportError:
            self.skipTest("Skipping test of faiss backend, faiss not found.")
        self.matching.n_neighbors = 3
        from causallib.contrib import faissknn
        def faiss_ivfflat(metric): return faissknn.FaissNearestNeighbors(index_type="ivfflat",metric=metric)
        def faiss_flatl2(metric): return faissknn.FaissNearestNeighbors(index_type="flatl2",metric=metric)
        for data in self.test_data:
            X, a, y = data

            def get_estimate_for_backend(backend, metric):
                self.matching.knn_backend = backend(metric) if callable(backend) else backend
                self.matching.metric = metric
                self.matching.fit(X, a, y)
                
                with warnings.catch_warnings():  # for some models this will throw a UserWarning
                    warnings.simplefilter(action="ignore", category=UserWarning)
                    Y = self.matching.estimate_population_outcome(X, a, y)
                return Y.values

            euclidean_results, mahalanobis_results = [[], []]
            for backend in ["sklearn", faiss_ivfflat, faiss_flatl2]:
                euclidean_results.append(
                    get_estimate_for_backend(backend, "euclidean"))
                mahalanobis_results.append(
                    get_estimate_for_backend(backend, "mahalanobis")
                )
            euclidean_results = np.array(euclidean_results)
            mahalanobis_results = np.array(mahalanobis_results)
            np.testing.assert_allclose(
                euclidean_results - euclidean_results[0], 0)
            np.testing.assert_allclose(
                mahalanobis_results - mahalanobis_results[0], 0)

    def _get_nearest_match_distances_for_treatment_value(self, treatment_value, a):

        return (self.matching.match_df_.loc[treatment_value][a != treatment_value]
                .distances.apply(lambda x: x[0] if x else np.nan)
                .dropna()
                .values)


    def test_is_pickleable(self):
        X, a, y = self.data_serial_unbalanced_x
        self.matching.fit(X,a,y)
        self.matching.match(X,a,)
        prepickle_covariates = self.matching.get_covariates_of_matches(0, 1, X)
        prepickle_estimate = self.matching.estimate_population_outcome(X, a).values
        ms = pickle.dumps(self.matching)

        m2 = pickle.loads(ms)
        postpickle_estimate = m2.estimate_population_outcome(X, a).values
        postpickle_covariates = m2.get_covariates_of_matches(0, 1, X)

        np.testing.assert_array_equal(
            prepickle_estimate, postpickle_estimate)
        np.testing.assert_array_equal(
            prepickle_covariates, postpickle_covariates)

    def test_matching_one_way_works_even_when_other_is_undefined(self):
        X, a, y = self.data_serial_unbalanced_x
        for n_neighbors in [5, 20, 50]:
            self.check_matching_too_few_neighbors_adapts_matches(n_neighbors, X, a, y)

    def check_matching_too_few_neighbors_adapts_matches(self, n_neighbors, X, a, y):
        matching = Matching(n_neighbors=n_neighbors, matching_mode="treatment_to_control")
        matching.fit(X,a,y)
        match_df = matching.match(X, a)
        n_matches_actual = match_df.distances.apply(len).groupby(level=0).max()
        self.assertEqual(n_matches_actual[0], min(n_neighbors, self.n))
        self.assertEqual(n_matches_actual[1], min(n_neighbors, self.k))
        


import unittest
import abc
from copy import deepcopy

import numpy as np
import pandas as pd
from sklearn.ensemble import GradientBoostingRegressor, GradientBoostingClassifier
from sklearn.linear_model import LinearRegression, LogisticRegression
from sklearn.model_selection import GridSearchCV, RandomizedSearchCV

from causallib.estimation import TMLE
from causallib.estimation import Standardization, IPW
from causallib.utils.general_tools import check_learner_is_fitted


def generate_data(n_samples, n_independent_features, n_interaction_features=None,
                  a_sparsity=0.8, y_sparsity=0.8,
                  X_normal=True,
                  seed=None):
    if seed is not None:
        np.random.seed(seed)

    # Generate independent features:
    if X_normal:
        X_independent = np.random.normal(size=(n_samples, n_independent_features))
    else:
        X_independent = np.random.binomial(1, 0.4, size=(n_samples, n_independent_features))
    X_independent = pd.DataFrame(
        X_independent,
        columns=[f"x_{i}" for i in range(X_independent.shape[1])]
    )

    # Generate treatment assignment:
    a_assignment, a_propensity, a_logit, a_beta = generate_vector(X_independent, a_sparsity)

    # Generate interactions:
    if n_interaction_features is None:
        n_interaction_features = n_independent_features
    assert n_interaction_features <= n_independent_features
    X_interactions = X_independent.sample(
        n=n_interaction_features,
        axis="columns",
    )
    X_interactions = X_interactions.multiply(a_assignment, axis="index")
    X_interactions = X_interactions.rename(columns=lambda s: f"{s}:a")
    # Generate outcome:
    X = X_independent.join(X_interactions)
    treatment_effect = 2
    y_binary, y_propensity, y_continuous, y_beta = generate_vector(X, y_sparsity, a_assignment, treatment_effect)

    data = {
        "X": X,
        "a": a_assignment,
        "y_cont": y_continuous,
        "y_bin": y_binary,
        "treatment_effect": treatment_effect,
        "y_beta": y_beta,
        "y_propensity": y_propensity,
    }
    return data


def generate_vector(X, sparsity, a=None, treatment_effect=None):
    beta = np.random.normal(size=X.shape[1])
    beta_mask = np.random.binomial(1, sparsity, size=beta.shape)
    beta_masked = beta * beta_mask
    logit = X @ beta_masked
    logit += np.random.normal(size=X.shape[0])  # Add noise
    if a is not None and treatment_effect is not None:
        logit += treatment_effect * a
    propensity = 1 / (1 + np.exp(-logit))
    classes = np.random.binomial(1, propensity)

    logit = pd.Series(logit, index=X.index)
    propensity = pd.Series(propensity, index=X.index)
    classes = pd.Series(classes, index=X.index)
    return classes, propensity, logit, beta_masked


class BaseTestTMLE(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        random_seed = 0
        cls.data = generate_data(200, 3, 3, seed=random_seed)

        cls.treatment_model = GradientBoostingClassifier()
        cls.outcome_model_bin = GradientBoostingClassifier()
        cls.outcome_model_cont = GradientBoostingRegressor()

    @abc.abstractmethod
    def init(self, reduced, importance_sampling):
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def estimator(self):
        raise NotImplementedError

    def ensure_fit(self):
        self.estimator.fit(
            self.data['X'], self.data['a'], self.data['y'],
        )
        self.assertTrue(check_learner_is_fitted(self.estimator.outcome_model.learner))
        self.assertTrue(check_learner_is_fitted(self.estimator.weight_model.learner))
        self.assertTrue(hasattr(self.estimator, "targeted_outcome_model_"))
        self.assertTrue(hasattr(self.estimator.targeted_outcome_model_, "params"))

    def ensure_estimate_individual_outcome(self):
        self.estimator.fit(
            self.data['X'], self.data['a'], self.data['y'],
        )
        ind_outcomes = self.estimator.estimate_individual_outcome(self.data['X'], self.data['a'])
        self.assertFalse(ind_outcomes.isna().any().any())

    def ensure_no_refit(self):
        n = self.data['X'].shape[0]
        self.estimator.fit(
            self.data['X'].loc[:n//2], self.data['a'].loc[:n//2], self.data['y'].loc[:n//2],
        )
        weight_model = deepcopy(self.estimator.weight_model)
        outcome_model = deepcopy(self.estimator.outcome_model)
        self.estimator.fit(
            self.data['X'].loc[n//2:], self.data['a'].loc[n//2:], self.data['y'].loc[n//2:],
            refit_weight_model=False
        )
        np.testing.assert_equal(
            weight_model.learner.feature_importances_,
            self.estimator.weight_model.learner.feature_importances_,
        )
        with self.assertRaises(AssertionError):  # Assert not equal since refitted
            np.testing.assert_equal(
                outcome_model.learner.feature_importances_,
                self.estimator.outcome_model.learner.feature_importances_,
            )


class BaseTestTMLEBinary(BaseTestTMLE):
    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        cls.data['y'] = cls.data['y_bin']

    def init(self, reduced, importance_sampling):
        self._estimator = TMLE(
            Standardization(self.outcome_model_bin, predict_proba=True),
            IPW(self.treatment_model),
            reduced=reduced, importance_sampling=importance_sampling,
        )

    @property
    def estimator(self):
        return self._estimator

    def ensure_warning_if_not_predict_proba(self):
        self.estimator.outcome_model.predict_proba = False
        with self.assertWarns(UserWarning):
            self.estimator.fit(self.data['X'], self.data['a'], self.data['y'])
        self.estimator.outcome_model.predict_proba = True

    def ensure_positive_label_prediction_is_used(self):
        self.estimator.fit(self.data['X'], self.data['a'], self.data['y'])
        tmle_output = self.estimator._outcome_model_estimate_individual_outcome(
            self.data['X'], self.data['a']
        )
        outcome_model_output = self.estimator.outcome_model.estimate_individual_outcome(
            self.data['X'], self.data['a']
        )
        outcome_model_output = outcome_model_output.loc[:, pd.IndexSlice[:, 1]]  # Take prediction for `1` class
        outcome_model_output = outcome_model_output.droplevel(level=-1, axis="columns")  # Drop redundant inner level
        pd.testing.assert_frame_equal(tmle_output, outcome_model_output)

    def ensure_average_effect(self):
        data = generate_data(1500, 2, 0, seed=0)
        self.estimator.fit(data['X'], data['a'], data['y_bin'])
        pop_outcome = self.estimator.estimate_population_outcome(data['X'], data['a'])
        effect = self.estimator.estimate_effect(pop_outcome[1], pop_outcome[0],
                                                effect_types=["diff", "ratio", "or"])
        np.testing.assert_allclose(
            data["y_propensity"][data['a'] == 1].mean() - data["y_propensity"][data['a'] == 0].mean(),
            effect['diff'],
            atol=0.1,
        )


class BaseTestTMLEContinuous(BaseTestTMLE):
    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        cls.data['y'] = cls.data['y_cont']

    def init(self, reduced, importance_sampling):
        self._estimator = TMLE(
            Standardization(self.outcome_model_cont),
            IPW(self.treatment_model),
            reduced=reduced, importance_sampling=importance_sampling,
        )

    @property
    def estimator(self):
        return self._estimator

    def ensure_target_scaling(self):
        self.estimator.fit(self.data['X'], self.data['a'], self.data['y'])
        with self.subTest("Test `target_scaler_` fit"):
            self.assertTrue(hasattr(self.estimator, "target_scaler_"))
            self.assertEqual(self.estimator.target_scaler_.data_min_[0], self.data['y'].min())
            self.assertEqual(self.estimator.target_scaler_.data_max_[0], self.data['y'].max())

        with self.subTest("Test `target_scaler_` inverse transform"):
            # Test predictions are not in a zero-one scale used for fitting
            ind_outcomes = self.estimator.estimate_individual_outcome(self.data['X'], self.data['a'])
            self.assertLess(ind_outcomes.min().min(), 0)
            self.assertLess(1, ind_outcomes.max().max())

    def ensure_average_effect(self):
        data = generate_data(1500, 2, 0, seed=0)
        self.estimator.fit(data['X'], data['a'], data['y_cont'])
        pop_outcome = self.estimator.estimate_population_outcome(data['X'], data['a'])
        effect = self.estimator.estimate_effect(pop_outcome[1], pop_outcome[0])
        # self.assertAlmostEqual(data['treatment_effect'], effect['diff'], places=1)
        np.testing.assert_allclose(
            data['treatment_effect'], effect['diff'],
            atol=0.1,
        )

    def ensure_conditional_effect(self):
        n_samples = 11000  # TODO: is it really that data inefficient to get within 0.1 of true parameters?
        data = generate_data(n_samples, 1, 1, a_sparsity=1.0, y_sparsity=1.0, X_normal=False, seed=0)
        self.estimator.fit(data['X'], data['a'], data['y_cont'])
        ind_outcome = self.estimator.estimate_individual_outcome(data['X'], data['a'])
        ind_effect = self.estimator.estimate_effect(ind_outcome[1], ind_outcome[0], agg="individual")
        ind_effect = ind_effect["diff"]

        # The modified effect should the added interaction term to the true effect
        np.testing.assert_allclose(
            ind_effect.loc[data['X'].iloc[:, -1] == 1].mean(),
            data['treatment_effect'] + data['y_beta'][-1],
            atol=0.6, rtol=1e-5,
        )

        # The effect under no modification should the coefficient of the treatment assignment
        np.testing.assert_allclose(
            ind_effect.loc[data['X'].iloc[:, -1] == 0].mean(),
            data['treatment_effect'],
            # decimal=1,
            atol=0.2, rtol=1e-5,
        )

        # The average effect should be the weighted mean between the two modifications
        np.testing.assert_almost_equal(
            ind_effect.mean(),
            np.average(
                [ind_effect.loc[data['X'].iloc[:, -1] == 0].mean(),
                 ind_effect.loc[data['X'].iloc[:, -1] == 1].mean()],
                weights=[sum(data['X'].iloc[:, -1] == 0),
                         sum(data['X'].iloc[:, -1] == 1)]
            ),
            decimal=3
        )


class TestTMLEMatrixFeatureBinary(BaseTestTMLEBinary):
    def setUp(self) -> None:
        self.init(reduced=False, importance_sampling=False)

    def test_fit(self):
        self.ensure_fit()

    def test_warning_if_not_predict_proba(self):
        self.ensure_warning_if_not_predict_proba()

    def test_positive_label_prediction_is_used(self):
        self.ensure_positive_label_prediction_is_used()

    def test_estimate_individual_outcome(self):
        self.ensure_estimate_individual_outcome()

    def test_no_refit(self):
        self.ensure_no_refit()

    def test_average_effect(self):
            self.ensure_average_effect()


class TestTMLEVectorFeatureBinary(BaseTestTMLEBinary):
    def setUp(self) -> None:
        self.init(reduced=True, importance_sampling=False)

    def test_fit(self):
        self.ensure_fit()

    def test_warning_if_not_predict_proba(self):
        self.ensure_warning_if_not_predict_proba()

    def test_positive_label_prediction_is_used(self):
        self.ensure_positive_label_prediction_is_used()

    def test_estimate_individual_outcome(self):
        self.ensure_estimate_individual_outcome()

    def test_raises_on_non_binary_treatment(self):
        data = deepcopy(self.data)
        data['a'].loc[:5] = 2  # Create multiclass treatment
        with self.assertRaises(AssertionError):
            self.estimator.fit(data['X'], data['a'], data['y'])

    def test_average_effect(self):
        self.ensure_average_effect()


class TestTMLEMatrixImportanceSamplingBinary(BaseTestTMLEBinary):
    def setUp(self) -> None:
        self.init(reduced=False, importance_sampling=True)

    def test_fit(self):
        self.ensure_fit()

    def test_warning_if_not_predict_proba(self):
        self.ensure_warning_if_not_predict_proba()

    def test_positive_label_prediction_is_used(self):
        self.ensure_positive_label_prediction_is_used()

    def test_estimate_individual_outcome(self):
        self.ensure_estimate_individual_outcome()

    def test_average_effect(self):
            self.ensure_average_effect()


class TestTMLEVectorImportanceSamplingBinary(BaseTestTMLEBinary):
    def setUp(self) -> None:
        self.init(reduced=True, importance_sampling=True)

    def test_fit(self):
        self.ensure_fit()

    def test_warning_if_not_predict_proba(self):
        self.ensure_warning_if_not_predict_proba()

    def test_positive_label_prediction_is_used(self):
        self.ensure_positive_label_prediction_is_used()

    def test_estimate_individual_outcome(self):
        self.ensure_estimate_individual_outcome()

    def test_raises_on_non_binary_treatment(self):
        data = deepcopy(self.data)
        data['a'].loc[:5] = 2  # Create multiclass treatment
        with self.assertRaises(AssertionError):
            self.estimator.fit(data['X'], data['a'], data['y'])

    def test_average_effect(self):
        self.ensure_average_effect()


class TestTMLEMatrixFeatureContinuous(BaseTestTMLEContinuous):
    def setUp(self) -> None:
        self.init(reduced=False, importance_sampling=False)

    def test_fit(self):
        self.ensure_fit()

    def test_target_scaling(self):
        self.ensure_target_scaling()

    def test_estimate_individual_outcome(self):
        self.ensure_estimate_individual_outcome()

    def test_average_effect(self):
        self.ensure_average_effect()

    def test_conditional_effect(self):
        self.ensure_conditional_effect()


class TestTMLEVectorFeatureContinuous(BaseTestTMLEContinuous):
    def setUp(self) -> None:
        self.init(reduced=True, importance_sampling=False)

    def test_fit(self):
        self.ensure_fit()

    def test_target_scaling(self):
        self.ensure_target_scaling()

    def test_estimate_individual_outcome(self):
        self.ensure_estimate_individual_outcome()

    def test_average_effect(self):
        self.ensure_average_effect()

    def test_conditional_effect(self):
        self.ensure_conditional_effect()


class TestTMLEMatrixImportanceSamplingContinuous(BaseTestTMLEContinuous):
    def setUp(self) -> None:
        self.init(reduced=False, importance_sampling=True)

    def test_fit(self):
        self.ensure_fit()

    def test_target_scaling(self):
        self.ensure_target_scaling()

    def test_estimate_individual_outcome(self):
        self.ensure_estimate_individual_outcome()

    def test_average_effect(self):
        self.ensure_average_effect()

    def test_conditional_effect(self):
        self.ensure_conditional_effect()


class TestTMLEVectorImportanceSamplingContinuous(BaseTestTMLEContinuous):
    def setUp(self) -> None:
        self.init(reduced=True, importance_sampling=True)

    def test_fit(self):
        self.ensure_fit()

    def test_target_scaling(self):
        self.ensure_target_scaling()

    def test_estimate_individual_outcome(self):
        self.ensure_estimate_individual_outcome()

    def test_average_effect(self):
        self.ensure_average_effect()

    def test_conditional_effect(self):
        self.ensure_conditional_effect()


from causallib.estimation.tmle import TargetMinMaxScaler


class TestTargetMinMaxScaler(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        cls.scaler = TargetMinMaxScaler(feature_range=(0, 1))
        cls.y = pd.Series(
            data=[-2, -1, 0, 1, 2],
            index=list("abcde"),
            name="target_mctargetface",
            dtype=float,
        )

    def setUp(self) -> None:
        self.scaler.fit(self.y)

    def test_fit(self):
        self.assertEqual(self.scaler.data_min_, -2)
        self.assertEqual(self.scaler.data_max_, 2)
        self.assertEqual(self.scaler.n_features_in_, 1)
        self.assertEqual(self.scaler.n_samples_seen_, 5)

    def test_transform(self):
        y = self.scaler.transform(self.y)
        self.assertEqual(y.min(), 0)
        self.assertEqual(y.max(), 1)
        self.assertIsInstance(y, pd.Series)
        self.assertEqual(y.name, self.y.name)
        pd.testing.assert_index_equal(y.index, self.y.index)

    def test_inverse_transform(self):
        y = self.scaler.transform(self.y)
        y = self.scaler.inverse_transform(y)
        pd.testing.assert_series_equal(y, self.y)



from .regression_curve_fitter import RegressionCurveFitter
from sklearn.base import BaseEstimator as SKLearnBaseEstimator
from typing import Any, Optional
import pandas as pd
from copy import deepcopy
from .survival_utils import canonize_dtypes_and_names, safe_join
from .base_survival import SurvivalBase


class StandardizedSurvival(SurvivalBase):
    def __init__(self,
                 survival_model: Any,
                 stratify: bool = True, **kwargs):
        """
        Standardization survival estimator.
        Computes parametric curve by fitting a time-varying hazards model that includes baseline covariates.
        Args:
            survival_model: Two alternatives:
                1. Scikit-Learn estimator (needs to implement `predict_proba`) - compute parametric curve by fitting a
                    time-varying hazards model that includes baseline covariates. Note that the model is fitted on a
                    person-time table with all covariates, and might be computationally and memory expansive.
                2. lifelines RegressionFitter - use lifelines fitter to compute survival curves from baseline covariates,
                    events and durations
            stratify (bool): if True, fit a separate model per treatment group
        """
        self.stratify = stratify

        if isinstance(survival_model, SKLearnBaseEstimator):
            # Construct default curve fitter, parametric with a scikit-learn estimator
            self.survival_model = RegressionCurveFitter(survival_model)
        else:
            # Initialized lifelines RegressionFitter (or any implementation with a compatible API)
            self.survival_model = survival_model

    def fit(self,
            X: pd.DataFrame,
            a: pd.Series,
            t: pd.Series,
            y: pd.Series,
            w: Optional[pd.Series] = None,
            fit_kwargs: Optional[dict] = None):
        """
        Fits parametric models and calculates internal survival functions.

        Args:
            X (pd.DataFrame): Baseline covariate matrix of size (num_subjects, num_features).
            a (pd.Series): Treatment assignment of size (num_subjects,).
            t (pd.Series): Followup duration, size (num_subjects,).
            y (pd.Series): Observed outcome (1) or right censoring event (0), size (num_subjects,).
            w (pd.Series): Optional subject weights.
            fit_kwargs (dict): Optional kwargs for fit call of survival model

        Returns:
            self
        """
        a, t, y, w, X = canonize_dtypes_and_names(a=a, t=t, y=y, w=w, X=X)
        if w is not None:
            fit_data, (w_name,) = safe_join(df=X, list_of_series=[w], return_series_names=True)
        else:
            fit_data = X
            w_name = None

        if fit_kwargs is None:
            fit_kwargs = {}

        self.stratified_curve_fitters_ = {}
        if self.stratify:
            fit_data, (t_name, y_name) = safe_join(df=fit_data, list_of_series=[t, y], return_series_names=True)
            unique_treatment_values = a.unique()
            for treatment_value in unique_treatment_values:
                stratum_curve_fitter = deepcopy(self.survival_model)
                stratum_curve_fitter.fit(df=fit_data[a == treatment_value], duration_col=t_name, event_col=y_name,
                                         weights_col=w_name, **fit_kwargs)
                self.stratified_curve_fitters_[treatment_value] = stratum_curve_fitter
        else:
            fit_data, (a_name, t_name, y_name) = safe_join(df=fit_data, list_of_series=[a, t, y],
                                                           return_series_names=True)
            self.survival_model.fit(df=fit_data, duration_col=t_name, event_col=y_name, weights_col=w_name,
                                    **fit_kwargs)

        return self

    def estimate_individual_outcome(self,
                                    X: pd.DataFrame,
                                    a: pd.Series,
                                    t: pd.Series,
                                    y: Optional[Any] = None,
                                    timeline_start: Optional[int] = None,
                                    timeline_end: Optional[int] = None
                                    ) -> pd.DataFrame:
        """
        Returns individual survival curves for each subject row in X/a/t

        Args:
            X (pd.DataFrame): Baseline covariate matrix of size (num_subjects, num_features).
            a (pd.Series): Treatment assignment of size (num_subjects,).
            t (pd.Series): Followup durations, size (num_subjects,).
            y: NOT USED (for API compatibility only).
            timeline_start (int): Common start time-step. If provided, will generate survival curves starting
                                  from 'timeline_start' for all patients. If None, will predict from first observed event (t.min()).
            timeline_end (int): Common end time-step. If provided, will generate survival curves up to 'timeline_end'
                                for all patients. If None, will predict up to last observed event (t.max()).

        Returns:
            pd.DataFrame: with time-step index, subject IDs (X.index) as columns and point survival as entries
        """

        min_time = timeline_start if timeline_start is not None else int(t.min())
        max_time = timeline_end if timeline_end is not None else int(t.max())
        contiguous_times = pd.Series(data=range(min_time, max_time + 1),
                                     name=t.name)  # contiguous time steps for inference

        a, _, _, _, X = canonize_dtypes_and_names(a=a, w=None, X=X)
        unique_treatment_values = a.unique()
        res = pd.DataFrame()
        for treatment_value in unique_treatment_values:
            if self.stratify:
                predict_data = X[a == treatment_value]
                stratum_curve_fitter = self.stratified_curve_fitters_[treatment_value]
            else:
                predict_data, a_name = safe_join(df=X, list_of_series=[a], return_series_names=True)
                stratum_curve_fitter = self.survival_model
            stratum_individual_survival_curves = stratum_curve_fitter.predict_survival_function(
                X=predict_data,
                times=contiguous_times
            )
            res = pd.concat([res, stratum_individual_survival_curves], axis=1)
        return res

    def estimate_population_outcome(self,
                                    X: pd.DataFrame,
                                    a: pd.Series,
                                    t: pd.Series,
                                    y: Optional[Any] = None,
                                    timeline_start: Optional[int] = None,
                                    timeline_end: Optional[int] = None
                                    ) -> pd.DataFrame:
        """
        Returns population averaged survival curves.

        Args:
            X (pd.DataFrame): Baseline covariate matrix of size (num_subjects, num_features).
            a (pd.Series): Treatment assignment of size (num_subjects,).
            t (pd.Series): Followup durations, size (num_subjects,).
            y: NOT USED (for API compatibility only).
            timeline_start (int): Common start time-step. If provided, will generate survival curves starting
                                  from 'timeline_start' for all patients. If None, will predict from first observed event (t.min()).
            timeline_end (int): Common end time-step. If provided, will generate survival curves up to 'timeline_end'
                                for all patients. If None, will predict up to last observed event (t.max()).

        Returns:
            pd.DataFrame: with time-step index, treatment values as columns and survival as entries
        """
        a, t, _, _, X = canonize_dtypes_and_names(a=a, t=t,  X=X)
        unique_treatment_values = a.unique()
        res = {}
        for treatment_value in unique_treatment_values:
            assignment = pd.Series(data=treatment_value, index=X.index, name=a.name)
            individual_survival_curves = self.estimate_individual_outcome(X=X, a=assignment, t=t,
                                                                          timeline_start=timeline_start,
                                                                          timeline_end=timeline_end)
            res[treatment_value] = individual_survival_curves.mean(axis='columns')

        res = pd.DataFrame(res)

        # Setting index/column names
        res.index.name = t.name
        res.columns.name = a.name
        return res



# (C) Copyright 2019 IBM Corp.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Created on Sept 25, 2019

import numpy as np
import pandas as pd
import torch

from causallib.estimation.base_estimator import IndividualOutcomeEstimator
from causallib.contrib.hemm.hemm import HEMM as HEMMTorch


class HEMM(IndividualOutcomeEstimator):
    """Causal model implementing an heterogeneous effect mixture model (HEMM). 

    For further details, see `Interpretable Subgroup Discovery in Treatment Effect 
    Estimation with Application to Opioid Prescribing Guidelines <https://arxiv.org/pdf/1905.03297.pdf>`_ paper.

    Notes:
        **Requires PyTorch** >= 1.2.0 (tested on 1.2.0)
    """

    def __init__(self,
                 D_in, K, homo=True, mu=None, std=None, bc=0, lamb=1e-4, spread=0.1, outcome_model='linear',
                 sep_heads=True, epochs=100, batch_size=10, learning_rate=1e-3, weight_decay=0.1, elbo_loss=True,
                 metric='AP', response='bin', use_p_correction=True, imb_fun='mmd2_lin', p_alpha=1e-4, random_seed=0):
        """Instantiate estimator.

        Args:
            D_in (int): Size of the features of the data
            K (int): Number of components to discover. (specifcy K-1: eg. For 2 components use K=1)
            homo (bool): Flag to specify if the final outcome model is same for each discovered subgroup.
                        Default is True ie. same outcome model is used for each subgroup.  
            mu (float): Initialize the components with means of the training data.
            std (float): Initialize the components with std dev of the training data.
            bc (int): The first feature in `x` being a bernoulli variables.
                    Columns 0 up to `bc` should be continuous (gaussian)
            lamb (float): Strength of the beta(0.5, 0.5) prior on the bernoulli variables.
            spread (float): How far should the components be initailized from there means.
            outcome_model (str): 'linear' to specify a linear outcome function.
                                 Or pass another Torch.model as the outcome model.
                                 Several outcome models are available in the `outcome_models.py` module.
            epochs (int): Max number of epochs.
            batch_size (int): Batch size for optimizer.
            learning_rate (float): Learning rate for the optimizer.
            weight_decay (float): Weight decay for the optimizer.
            elbo_loss (bool): Wether to use evidence lower bound (ELBO) as the loss function.
            metric (str): 'AP' mean average precision, 'AuROC': area under roc curve, 'MSE':mean squared error.
                          If not specified it uses the optimizer cost as the metric.
                          The specified metric is computed for the training set (or dev set if it is specified in `fit`)
                          and used to perform early stopping to prevent overfitting if dev set is provided.
            response (str): Specify if y is continuous ('cont') or binary ('bin').
            use_p_correction (bool): Whether to use population size p(treated) in imbalance penalty (IPM).
            imb_fun (str): Which imbalance penalty to use ('mmd2_lin', 'wass').
            p_alpha (float): Imbalance regularization parameter.
            random_seed (int): Random Seed to initialize model parameters.
        """
        self._epochs = epochs
        self._batch_size = batch_size
        self._learning_rate = learning_rate
        self._weight_decay = weight_decay
        self._elbo_loss = elbo_loss
        self._metric = metric
        self._response = response
        self._use_p_correction = use_p_correction
        self._imb_fun = imb_fun
        self._p_alpha = p_alpha
        self._sep_heads = sep_heads
        self._homo = homo
        self._random_seed = random_seed
        
        torch.manual_seed(random_seed)

        model = HEMMTorch(D_in, K, homo, mu, std, bc, lamb, spread=spread, outcomeModel=outcome_model, sep_heads=sep_heads)
        model = model.double()
        super(HEMM, self).__init__(learner=model)

    def estimate_individual_outcome(self, X, a, treatment_values=None, soft=True):
        """
        Estimates individual outcome under different treatment values (interventions)

        Args:
            X (torch.Tensor | pd.DataFrame | np.ndarray): Covariate matrix of size (num_subjects, num_features).
            a (torch.Tensor | pd.Series | np.ndarray): Treatment assignment of size (num_subjects,).
            treatment_values: subset of values to get potential outcome for.
                              If None return for all unique treatment_values found in `a`.
            soft (bool):

        Returns:
            pd.DataFrame: DataFrame which columns are treatment values and rows are individuals: each column is a vector
                          size (num_samples,) that contains the estimated outcome for each individual under the
                          treatment value in the corresponding key.
        """
        ys = self.learner.estimate_individual_outcomes(self._as_tensor(X), self._as_tensor(a),
                                                       response=self._response, soft=soft)
        ys = pd.DataFrame(zip(*[y.cpu().data.numpy() for y in ys]))
        if treatment_values is not None:
            ys = ys[treatment_values]
        return ys

    def get_groups(self, X):
        """Return hard assignment of groups for each sample

        Args:
            X (torch.Tensor | pd.DataFrame | np.ndarray): Covariate matrix of size (num_subjects, num_features).

        Returns:
            groups (pd.Series): Most probable group assignment of each sample, size = (num_samples,)
        """
        groups = self.learner.get_groups(self._as_tensor(X))
        groups = pd.Series(groups)
        return groups

    def get_groups_proba(self, X, log=False):
        """Return soft assignment of probability of each sample to be part of each group.

        Args:
            X (torch.Tensor): Covariate matrix of size (num_subjects, num_features).
            log (bool): If True returns log probabilities

        Returns:
            z_pred (np.array): probability of group membership given X P(Z|X),
                               size = (num_covariates, num_components+1).
        """
        X = self._as_tensor(X)
        groups = self.learner.get_groups_proba(X, log=log)
        groups = pd.DataFrame(groups)
        return groups

    def get_groups_effect(self, X=None, a=None):
        """Returns effect (\\gamma_k in the paper) in each group as estimated during training"""
        group_effect = self.learner.get_groups_effect()
        return group_effect

    def group_sizes(self, X):
        """Return underlying treatment groups sizes.
        Args:
            X (torch.Tensor | pd.DataFrame | np.ndarray): Covariate matrix of size (num_subjects, num_features).

        Returns: collection.Counter giving size for each group
        """
        return self.learner.group_sizes(self._as_tensor(X))

    def fit(self, X, a, y, sample_weight=None, validation_data=None):
        """Trains a causal model from observed data.

        Note: If all dev-data is provided, early stopping criteria will be applied using this data and the metric
              specified in self._metric.

        Args:
            X (torch.Tensor | pd.DataFrame | np.ndarray): Covariate matrix of size (num_subjects, num_features).
            a (torch.Tensor | pd.Series | np.ndarray): Treatment assignment of size (num_subjects,).
            y (torch.Tensor | pd.Series | np.ndarray): Observed outcome of size (num_subjects,).
            sample_weight: *IGNORED*
            validation_data: tuple of validation set: (X_val, a_val, y_val) corresponding to `X, a, y` above.

        Returns:
            IndividualOutcomeEstimator: A causal weight model with an inner learner fitted.
        """
        ltype = 'log' if self._elbo_loss else None
        if validation_data is not None:
            validation_data = (self._as_tensor(d) for d in validation_data)
        else:
            validation_data = None

        return self.learner.fit(
            train=(self._as_tensor(X), self._as_tensor(a), self._as_tensor(y)),
            epochs=self._epochs,
            batch_size=self._batch_size,
            lr=self._learning_rate,
            wd=self._weight_decay,
            ltype=ltype,
            dev=validation_data,
            metric=self._metric,
            response=self._response,
            use_p_correction=self._use_p_correction,
            imb_fun=self._imb_fun,
            p_alpha=self._p_alpha
        )

    @staticmethod
    def _as_tensor(X):
        """Convert the given argument to a Torch tensor.
        
            Args:
                X (pd.DataFrame|pd.Series|np.ndarray|torch.Tensor): argument to convert
            Returns:
                torch.Tensor with the same shape as the given argument
            Raises:
                TypeError: if X is not a DataFrame/Series, Numpy array, or Torch tensor
        """
        if isinstance(X, pd.DataFrame) or isinstance(X, pd.Series):
            return torch.from_numpy(X.values.astype('float64'))
        if isinstance(X, np.ndarray):
            return torch.from_numpy(X.astype('float64'))
        if isinstance(X, torch.Tensor):
            return X
        raise TypeError('Argument must be Numpy array, Panda DataFrame/Series, or PyTorch tensor')



from sklearn.model_selection import KFold, StratifiedKFold
from sklearn.utils.metaestimators import _safe_split
import pandas as pd
from sklearn.base import clone


def cross_fitting(estimator, X, y, n_splits=5, predict_proba=False,
                  return_estimator=True):
    """

    Args:
        estimator(object): sklearn object
        X (pd.DataFrame): Covariate matrix of size (num_subjects, num_features).
        y (pd.Series): Observed outcome of size (num_subjects,).
        n_splits (int): number of folds
        predict_proba (bool): If True, the treatment model is a classifier
                                and use 'predict_proba',
                              If False, use 'predict'.
        return_estimator (bool): If true return fitted estimators of each fold

    Returns:
        array of held-out prediction,
        if return estimator:
            a tuple of estimators on held-out-data
    """

    cv = StratifiedKFold(n_splits=n_splits) if predict_proba else KFold(
        n_splits=n_splits)
    ret = [_fit_and_predict(clone(estimator), X, y, train, test,
                            predict_proba=predict_proba)
           for train, test in cv.split(X, y)]
    zipped_ret = list(zip(*ret))
    if return_estimator:
        return pd.concat(zipped_ret[0]), zipped_ret[1]
    else:
        return pd.concat(zipped_ret[0])


def _fit_and_predict(estimator, X, y, train, test, predict_proba):
    """
    fit the estimator with the train samples and make prediction with the test data
    Args:
        estimator(object): sklearn object
        X (pd.DataFrame): Covariate matrix of size (num_subjects, num_features).
        y (pd.Series): Observed outcome of size (num_subjects,).
        train:
        test:
        predict_proba (bool): If True, the treatment model is a classifier
                                and use 'predict_proba',
                              If False, use 'predict'.

    """
    X_train, y_train = _safe_split(estimator, X, y, train)
    X_test, _ = _safe_split(estimator, X, y, test, train)
    estimator.fit(X_train, y_train)
    if predict_proba:
        pred = estimator.predict_proba(X_test)[:, 1]
    else:
        pred = estimator.predict(X_test)

    return pd.Series(pred, index=X_test.index), estimator



"""Predictions from single folds.

Predictions are generated by predictors for causal models. They contain the estimates
for single folds and are combined in the EvaluationResults objects for further analysis.
"""
from collections import namedtuple
from typing import Union
import warnings

import pandas as pd
from ..utils.stat_utils import robust_lookup
from .metrics import evaluate_metrics
from ..metrics.weight_metrics import calculate_covariate_balance


WeightEvaluatorScores = namedtuple(
    "WeightEvaluatorScores", ["prediction_scores", "covariate_balance"]
)


class WeightPredictions:
    """Data structure to hold weight-model predictions"""

    def __init__(
        self,
        weight_by_treatment_assignment,
        weight_for_being_treated,
        treatment_assignment_pred,
    ):
        self.weight_by_treatment_assignment = weight_by_treatment_assignment
        self.weight_for_being_treated = weight_for_being_treated
        self.treatment_assignment_pred = treatment_assignment_pred

    def evaluate_metrics(self, X, a_true, metrics_to_evaluate):
        """
        Evaluate metrics on prediction.

        Args:
            X (pd.DataFrame): Covariates.
            a_true (pd.Series): ground truth treatment assignment
            metrics_to_evaluate (dict | None): key: metric's name, value: callable that receives
                true labels, prediction and sample_weights (the latter may be ignored).

        Returns:
            WeightEvaluatorScores: Object with two data attributes: "predictions"
                and "covariate_balance"
        """
        y_pred_proba, y_pred = self._get_predictions_to_evaluate()
        evaluated_metrics = evaluate_metrics(
            metrics_to_evaluate=metrics_to_evaluate,
            y_true=a_true,
            y_pred_proba=y_pred_proba,
            y_pred=y_pred,
        )
        # Convert single-dtype Series to a row in a DataFrame:
        evaluated_metrics_df = pd.DataFrame(evaluated_metrics).T
        # change dtype of each column to numerical if possible:
        evaluated_metrics_df = evaluated_metrics_df.apply(
            pd.to_numeric, errors="ignore"
        )

        covariate_balance = calculate_covariate_balance(
            X, a_true, self.weight_by_treatment_assignment
        )

        results = WeightEvaluatorScores(evaluated_metrics_df, covariate_balance)
        return results

    def _get_predictions_to_evaluate(self):
        y_pred_proba = self.weight_for_being_treated
        y_pred = self.treatment_assignment_pred
        return y_pred_proba, y_pred


class PropensityPredictions(WeightPredictions):
    """Data structure to hold propensity-model predictions"""

    def __init__(
        self,
        weight_by_treatment_assignment,
        weight_for_being_treated,
        treatment_assignment_pred,
        propensity,
        propensity_by_treatment_assignment,
    ):
        super().__init__(
            weight_by_treatment_assignment,
            weight_for_being_treated,
            treatment_assignment_pred,
        )
        self.propensity = propensity
        self.propensity_by_treatment_assignment = propensity_by_treatment_assignment

    def _get_predictions_to_evaluate(self):
        y_pred_proba = self.propensity
        y_pred = self.treatment_assignment_pred
        return y_pred_proba, y_pred

class OutcomePredictions:
    """Data structure to hold outcome-model predictions"""

    def __init__(self, prediction, prediction_event_prob=None):
        self.prediction = prediction
        self.prediction_event_prob = self._correct_predict_proba_estimate(
            prediction, prediction_event_prob
        )
        self.is_binary_outcome = self.prediction_event_prob is not None

    @staticmethod
    def _correct_predict_proba_estimate(prediction, prediction_event_prob):
        # Estimation output for predict_proba=True has same columns as for predict_proba=False.
        # This means either base-learner has no predict_proba/decision_function
        # or problem is not classification.
        # Either way, it means there are no prediction probabilities
        if prediction_event_prob.columns.tolist() == prediction.columns.tolist():
            return None

        # predict_proba=True was able to predict probabilities. However,
        # Prediction probability evaluation is only applicable for binary outcome:
        y_values = prediction_event_prob.columns.get_level_values("y").unique()
        # Note: on pandas 23.0.0 you could do prediction_event_prob.columns.unique(level='y')
        if y_values.size == 2:
            event_value = y_values.max()
            # get the maximal value, assumes binary 0-1 (1: event, 0: non-event)
            # Extract the probability for event:
            return prediction_event_prob.xs(key=event_value, axis="columns", level="y")

        warnings.warn(
            "Multiclass probabilities are not well defined and supported for evaluation.\n"
            "Falling back to class predictions.\n"
            "Plots might be uninformative due to input being classes and not probabilities."
        )
        return None

    def evaluate_metrics(self, a, y, metrics_to_evaluate):
        """Evaluate metrics for this model prediction.

        Args:
            a (pd.Series): treatment assignment
            y (pd.Series): ground truth outcomes
            metrics_to_evaluate (Dict[str,Callable]): key: metric's name, value: callable that
                receives true labels, prediction and sample_weights (the latter may be ignored).
                If not provided, defaults from causallib.evaluation.metrics are used.

        Returns:
            pd.DataFrame: evaluated metrics
        """

        scores = {"actual": self._evaluate_metrics_overall(a, y, metrics_to_evaluate)}

        scores.update(
            {
                str(t): self._evaluate_metrics_on_treatment_value(
                    a, y, metrics_to_evaluate, t
                )
                for t in sorted(set(a))
            }
        )

        scores = pd.concat(scores, names=["model_strata"], axis="columns").T
        scores = scores.apply(pd.to_numeric, errors="ignore")
        return scores

    def _evaluate_metrics_on_treatment_value(
        self, a_true, y_true, metrics_to_evaluate, treatment_value
    ):
        # Stratify based on treatment assignment:
        y_is_binary = y_true.nunique() == 2
        treatment_value_idx = a_true == treatment_value
        y_true_strata = y_true.loc[treatment_value_idx]
        prediction_strata = self.prediction.loc[treatment_value_idx, treatment_value]
        if y_is_binary:
            prediction_prob_strata = self.prediction_event_prob.loc[
                treatment_value_idx, treatment_value
            ]
        else:
            prediction_prob_strata = None

        evaluated_metrics = evaluate_metrics(
            metrics_to_evaluate=metrics_to_evaluate,
            y_true=y_true_strata,
            y_pred=prediction_strata,
            y_pred_proba=prediction_prob_strata,
        )

        return evaluated_metrics

    def _evaluate_metrics_overall(self, a_true, y_true, metrics_to_evaluate):
        # Score overall:
        # # Extract prediction on actual treatment
        y_is_binary = y_true.nunique() == 2
        prediction_strata = robust_lookup(self.prediction, a_true)
        if y_is_binary:
            prediction_prob_strata = robust_lookup(self.prediction_event_prob, a_true)
        else:
            prediction_prob_strata = None

        evaluated_metrics = evaluate_metrics(
            metrics_to_evaluate=metrics_to_evaluate,
            y_true=y_true,
            y_pred=prediction_strata,
            y_pred_proba=prediction_prob_strata,
        )

        return evaluated_metrics

    def get_prediction_by_treatment(self, a: pd.Series):
        """Get proba if available else prediction"""
        if self.is_binary_outcome:
            pred = self.prediction_event_prob
        else:
            pred = self.prediction
        return robust_lookup(pred, a[pred.index])

    def get_proba_by_treatment(self, a: pd.Series):
        """Get proba of prediction"""
        return robust_lookup(self.prediction_event_prob, a[self.prediction.index])


SingleFoldPrediction = Union[
    PropensityPredictions, WeightPredictions, OutcomePredictions
]
